package learningRules;
// BGLearning for a n state n strategy model.


import java.util.*;

import utilities.Game;
import utilities.Individual;
import utilities.ModelError;


/**
* This is a strategy learning model for the Bala Goyal Model.  This model presumes there are 
* n states of the world and n strategies.  It then has the individuals update probabilites base
* on the state of the world.
* @author Kevin Zollman
*/

public class BgLearning implements LearningRule {
        
        Game games[];
        
        /**
        * This does nothing since there is no discounting or anything.
        * @param i The individual calling the function
        * @throws ModelError (although, since this does nothing it will never be thrown)
        */
        public void newGeneration(Individual i) throws ModelError { }
        
        /**
        * This processes a payoff for an individual.  This function performs standard Bayesian 
        * updating based the the most recent payoff received This function should only be called 
        * after the game has been played.
        * @param i The individual calling the function
        * @throws ModelError if something fails durring updating
        */
        public synchronized void processPayoff(Individual i) throws ModelError { 
                double denom = 0.0;
                double[] newBeliefs;
                
                // first based on your payoff
                for(int k=0;k<games.length;k++) 
                        denom += (double) games[k].getProb(i.getStrat(), i.getPayoff())*i.getBelief(k);
                
                newBeliefs = new double[games.length];
                for(int k=0;k<games.length;k++) {
                        newBeliefs[k] = (double) ( (double) games[k].getProb(i.getStrat(), i.getPayoff()) * i.getBelief(k))/denom;
                        if(Double.isNaN(newBeliefs[k]))
                                throw new ModelError("NaN Belief: " + games[k].getProb(i.getStrat(), i.getPayoff()) +"*" + i.getBelief(k) + " / " + denom);
                }
                        
                //then your neighbors
                HashSet neigh = i.getConnected();
                synchronized(neigh) {
                Iterator j = neigh.iterator();
                while(j.hasNext()) {
                        Individual n = (Individual) j.next();
                        
                        denom = 0.0;
                        for(int k=0;k<games.length;k++) 
                                denom += (double) games[k].getProb(n.getStrat(), n.getPayoff())*newBeliefs[k];
                        for(int k=0;k<games.length;k++) {
                                newBeliefs[k] = (double) ((double) games[k].getProb(n.getStrat(), n.getPayoff()) * newBeliefs[k])/denom;
                                if(Double.isNaN(newBeliefs[k]))
                                        throw new ModelError("NaN Belief (" +i.getId() + "/" + n.getId() + "): " + games[k].getProb(n.getStrat(), n.getPayoff()) +"*" + newBeliefs[k] +" / " +denom);
                        }
                }
                }
                
                i.setBeliefs(newBeliefs);
                
        }
        
        /**
        * This function chooses a strategy for the player.  It uses a simple best response to its
        * beliefs function.  The best response is deterministic.
        * @param i The calling individual
        */
        public synchronized int pickStrat(Individual i) {
                double expectation[] = new double[games[0].getStrats()];
                
                for(int j=0;j<games[0].getStrats();j++)
                        expectation[j] = 0.0;
                for(int j=0;j<games[0].getStrats();j++) {
                        for(int k=0;k<games.length;k++) 
                               expectation[j] += (double) i.getBelief(k)*games[k].getExpectedPayoff(j);
                }
                
                double highestexpectation = 0.0;
                int strat = 0;
                for(int j=0;j<expectation.length;j++) {
                        if (expectation[j] > highestexpectation) {
                                strat = j;
                                highestexpectation = expectation[j];
                        }
                }
                
                return strat;
        }
        
        /**
        * Initializing a BgLearning model.
        * @param g The games array containing all the possible games one for each state of the 
        *          world
        */
        public BgLearning(Game g[]) { games = g;}
        
}
