package learningRules;

import java.util.*;

import utilities.Game;
import utilities.Individual;
import utilities.ModelError;


/**
* This class implements myopic best response learning.  Here an individual looks 
* at the strategies employed by her neighbors on the last round (only) and chooses
* the strategy on the next round which was the best response to their action on the
* last round.
* @Author Kevin Zollman
*/
public class MyopicBRLearning implements LearningRule {
        
        Game game;
        
        /**
        * This does nothing since there is no discounting or anything.
        * @param i The individual calling the function
        * @throws ModelError (although, since this does nothing it will never be thrown)
        */
        public void newGeneration(Individual i) throws ModelError { }
        
        /**
        * Processing a payoff does nothing, since there are no beliefs to be updated.
        * @param i The individual calling the function
        * @throws ModelError (although, since this does nothing it will never be thrown)
        */
        public void processPayoff(Individual i) throws ModelError { }
        
        /**
        * This function chooses a strategy which is a best response to the collective action
        * of the players last round.  If the players previous strategy is tied with another
        * as the best, the player keeps the strategy.  If two strategies (not used by the
        * player on the last round) are tied, the player chooses one at random.
        * @param i The calling individual
        */
        public int pickStrat(Individual i) {
                //gets a list of neighbors actions
                int[] oppstratcount = new int[game.getStrats()];
                for(int j=0;j<oppstratcount.length;j++)
                        oppstratcount[j] = 0;
                                
                HashSet neigh = i.getConnected();
                synchronized(neigh) {
                Iterator j = neigh.iterator();
                while(j.hasNext()) {
                        Individual n = (Individual) j.next();
                        
                        oppstratcount[n.getLastStrat()]++;

                }
                }
                
                int defstrat = i.getLastStrat();
                int defstratpayoff = 0;
                if(game.getStochastic()) {
                        for(int k=0;k<oppstratcount.length;k++) {
                                defstratpayoff += oppstratcount[k] *
                                        game.getExpectedPayoff(defstrat,k);
                        }
                }
                else {
                        defstratpayoff = i.getLastPayoff();
                }
                
                int[] stratpayoffs = new int[oppstratcount.length]; 
                
                int bestpayoff = defstratpayoff;
                int[] beststrats = new int[oppstratcount.length];
                int numbeststrats = 1;
                beststrats[0] = defstrat;
                
                for(int m=0;m<oppstratcount.length;m++) {
                        if(m != defstratpayoff) {
                                stratpayoffs[m] = 0;
                                for(int k=0;k<oppstratcount.length;k++) {
                                        stratpayoffs[m] += oppstratcount[k] *
                                                game.getExpectedPayoff(m,k);
                                }
                                if(stratpayoffs[m]>defstratpayoff && stratpayoffs[m]>=bestpayoff) {
                                        if(stratpayoffs[m]==bestpayoff) {
                                                beststrats[numbeststrats] = m;
                                                numbeststrats++;
                                        }
                                        else {
                                                bestpayoff=stratpayoffs[m];
                                                beststrats[0] = m;
                                                numbeststrats = 1;
                                        }
                                }
                        }
                }
                
                return beststrats[i.getRandom().nextInt(numbeststrats)];
        }
        
        /**
        * Constructor
        * @param g The game
        */
        public MyopicBRLearning(Game g) { game = g; }
}
