package learningRules;
// SmoothBGLearning for a n state n strategy model.


import java.util.*;

import utilities.Game;
import utilities.Individual;
import utilities.ModelError;

/**
* This is a strategy learning model for the Bala Goyal Model.  This model presumes there are 
* n states of the world and n strategies.  It then has the individuals update probabilites based
* on the state of the world.  The best response rule is a probabilistic (smoothed) best response.
* @author Kevin Zollman
*/

public class SmoothBGLearning implements LearningRule {
        
        Game games[];
        double gamma;
        
        /**
        * This calculates 1/(e^x[0]+e^x[1]+...+e^x[n])
        * It returns a double which is a rounded version of the result
        * @param x The list of exponents
        * @return 1/(e^x[0]+e^x[1]+...+e^x[n])
        */
        public static final double probs(double[] x) {
                double total = 0.0;
                for(int i=0;i<x.length;i++) {
                        double subt = Math.exp(x[i]);
                        total += subt;
                }
                return (double) 1/total;
        }
        
        /**
        * This does nothing since there is no discounting or anything.
        * @param i The individual calling the function
        * @throws ModelError (although, since this does nothing it will never be thrown)
        */
        public void newGeneration(Individual i) throws ModelError { }
        
        /**
        * This processes a payoff for an individual.  This function performs standard Bayesian 
        * updating based the the most recent payoff received This function should only be called 
        * after the game has been played.
        * @param i The individual calling the function
        * @throws ModelError if something fails durring updating
        */
        public synchronized void processPayoff(Individual i) throws ModelError { 
                double denom = 0.0;
                double[] newBeliefs;
                
                // first based on your payoff
                for(int k=0;k<games.length;k++) 
                        denom += (double) games[k].getProb(i.getStrat(), i.getPayoff())*i.getBelief(k);
                
                newBeliefs = new double[games.length];
                for(int k=0;k<games.length;k++) {
                        newBeliefs[k] = (double) ( (double) games[k].getProb(i.getStrat(), i.getPayoff()) * i.getBelief(k))/denom;
                        if(Double.isNaN(newBeliefs[k]))
                                throw new ModelError("NaN Belief: " + games[k].getProb(i.getStrat(), i.getPayoff()) +"*" + i.getBelief(k) + " / " + denom);
                }
                        
                //then your neighbors
                HashSet neigh = i.getConnected();
                synchronized(neigh) {
                Iterator j = neigh.iterator();
                while(j.hasNext()) {
                        Individual n = (Individual) j.next();
                        
                        denom = 0.0;
                        for(int k=0;k<games.length;k++) 
                                denom += (double) games[k].getProb(n.getStrat(), n.getPayoff())*newBeliefs[k];
                        for(int k=0;k<games.length;k++) {
                                newBeliefs[k] = (double) ((double) games[k].getProb(n.getStrat(), n.getPayoff()) * newBeliefs[k])/denom;
                                if(Double.isNaN(newBeliefs[k]))
                                        throw new ModelError("NaN Belief (" +i.getId() + "/" + n.getId() + "): " + games[k].getProb(n.getStrat(), n.getPayoff()) +"*" + newBeliefs[k] +" / " +denom);
                        }
                }
                }
                
                i.setBeliefs(newBeliefs);
        }
        
        /**
        * This function chooses a strategy for the player.  It uses a probabilistic (smoothed) 
        * best response to its beliefs function.  The function used is an exponentiation function
        * for two strategies the probability of strategy one is 
        * <pre>e^(E(1)/gamma) / *(e^(E(1)/gamma) + e^(E(2)/gamma))</pre> where E(x) is the 
        * expectation of action x and gamma is a parameter which corresponds to the amount
        * of "smoothing" that occurs (i.e. the higher the gamma the further from the 
        * best response correspondence the function is).
        * @param i The calling individual
        */
        public synchronized int pickStrat(Individual i){
                double expectation[] = new double[games[0].getStrats()];
                
                //Calculate the expectation for each action
                for(int j=0;j<expectation.length;j++)
                        expectation[j] = 0.0;
                for(int j=0;j<expectation.length;j++) {
                        for(int k=0;k<games.length;k++) 
                               expectation[j] += (double) i.getBelief(k)*games[k].getExpectedPayoff(j);
                }
                
                //Divide by gamma
                for(int j=0;j<expectation.length;j++) 
                        expectation[j] = expectation[j]/gamma;
                
                // Assign probabilities to each action using the exponentiation best response
                //    rule
                // This code looks strange, but this is the most optmized I could come up with
                
                double[] probabilities = new double[expectation.length];
                
                probabilities[expectation.length-1] = 1.0;
                double[] diffs = new double[expectation.length];
                for(int j=0;j<(expectation.length-1);j++) {
                        for(int k=0;k<expectation.length;k++) {
                                diffs[k] = (double) expectation[k] - expectation[j];
                        }
                        probabilities[j] = SmoothBGLearning.probs(diffs);
                        probabilities[expectation.length-1] -= probabilities[j];
                }
                
                //Figure out which action to take on this round
                double number = i.getRandom().nextDouble();
                double count = probabilities[0];
                for(int j=0;j<expectation.length;j++) { 
                        if(number <= count)
                                return j;
                        else
                                count += probabilities[j+1];
                }
                
                //Shouldn't ever get here
                return -1;
         }
        
        /**
        * Initializing a BgLearning model.
        * @param g The games array containing all the possible games one for each state of the 
        *          world
        */
        public SmoothBGLearning(Game g[], double gam) { games = g; gamma = gam;  }
        
}
