package learningRules;
import utilities.Individual;
import utilities.ModelError;

/**
* This implements a modification of the standard reinforcement dynamics.
*/

public class SmoothReinforcement implements LearningRule {
        
        /**
        * This calculates 1/(e^x[0]+e^x[1]+...+e^x[n])
        * It returns a double which is a rounded version of the result
        * @param x The list of exponents
        * @return 1/(e^x[0]+e^x[1]+...+e^x[n])
        */
        public static final double probs(double[] x) {
                double total = 0.0;
                for(int i=0;i<x.length;i++) {
                        double subt = Math.exp(x[i]);
                        total += subt;
                }
                return (double) 1/total;
        }
        
        double delta;	
        /**
	 * Returns the value of delta.
	 */
	public double getDelta()
	{
		return delta;
	}

	/**
	 * Sets the value of delta.
	 * @param delta The value to assign delta.
	 */
	public void setDelta(double delta)
	{
		this.delta = delta;
	}

        
        double beta;	
        /**
	 * Returns the value of beta.
	 */
	public double getBeta()
	{
		return beta;
	}

	/**
	 * Sets the value of beta.
	 * @param beta The value to assign beta.
	 */
	public void setBeta(double beta)
	{
		this.beta = beta;
	}
        
        /**
        * Prepares the individual for a new generation.  This does nothing since the discount
        *  factor is already applied in the updating procedure.
        * @param i The calling individual
        */
        public synchronized void newGeneration(Individual i) throws ModelError { }
         
      

        /**
        * Processes the payoff by adding the payoff to the weight of the strategy choosen on this
        * round
        * @param i The calling individual
        */
        public synchronized void processPayoff(Individual i) throws ModelError {
                i.setStratWeights(i.getStrat(), 
                        (double) ((double) (1-delta) * i.getWeights(i.getStrat())) + 
                                ((double) delta * i.getPayoff()));
        }

        /**
        * Processes the payoff by adding the payoff to the weight of the strategy choosen on this
        * round.  This is for multi-domain models
        * @param i The calling individual
        */
        public synchronized void processPayoff(Individual i, int d) throws ModelError {
                if(i.getPayoff() > 0) {
                        i.setStratWeights(d, i.getStrat(), 
                                (double) ((double) (1-delta) * i.getWeights(d, i.getStrat())) + 
                                        ((double) delta * i.getPayoff()) );
                }
        }

        
        /** 
        * Picks a strategy based on a logisitic response rule similar to smoothed fictitious play
        * this is for multi domain models
        * @param i The calling individual
        */
        public synchronized int pickStrat(Individual i){
                
                double[] weights = i.getWeights();
                
                // Assign probabilities to each action using the exponentiation best response
                //    rule
                // This code looks strange, but this is the most optmized I could come up with
                
                double[] probabilities = new double[weights.length];
                
                //Save calculating one probability
                probabilities[weights.length-1] = 1.0;
                double[] diffs = new double[weights.length];
                for(int j=0;j<(weights.length -1);j++) {
                        for(int k=0;k<weights.length;k++) {
                                diffs[k] = (double) beta * ((double) weights[k] - weights[j]);
                        }
                        probabilities[j] = SmoothReinforcement.probs(diffs);
                        probabilities[weights.length-1] -= probabilities[j];
                }
                
                //Figure out which action to take on this round
                double number = i.getRandom().nextDouble();
                double count = probabilities[0];
                for(int j=0;j<weights.length;j++) { 
                        if(number <= count)
                                return j;
                        else
                                count += probabilities[j+1];
                }
                
                //Shouldn't ever get here
                return -1;
        }
       
        /** 
        * Picks a strategy based on a logisitic response rule similar to smoothed fictitious play
        * this is for signaling games
        * @param i The calling individual
        */
        public synchronized int pickStrat(Individual i, int d){
                
                double[] weights = i.getDomainWeights(d);
                
                // Assign probabilities to each action using the exponentiation best response
                //    rule
                // This code looks strange, but this is the most optmized I could come up with
                
                double[] probabilities = new double[weights.length];
                
                //Save calculating one probability
                probabilities[weights.length-1] = 1.0;
                double[] diffs = new double[weights.length];
                for(int j=0;j<(weights.length -1);j++) {
                        for(int k=0;k<weights.length;k++) {
                                diffs[k] = (double) beta * ((double) weights[k] - weights[j]);
                        }
                        probabilities[j] = SmoothReinforcement.probs(diffs);
                        probabilities[weights.length-1] -= probabilities[j];
                }
                
                try {
                //Figure out which action to take on this round
                double number = i.getRandom().nextDouble();
                double count = probabilities[0];
                for(int j=0;j<weights.length;j++) { 
                        if(number <= count)
                                return j;
                        else
                                count += probabilities[j+1];
                }
                
                }
                catch (java.lang.ArrayIndexOutOfBoundsException e) {
                        System.out.println(beta + " " + delta);
                        for(int j=0;j<weights.length;j++) {
                                System.out.print(i.getWeights(d, j) + " ");
                                System.out.print(weights[j] + " ");
                                System.out.println(probabilities[j]);
                        }
                        System.out.println("Array index of of bounds");
                }
                //Shouldn't ever get here
                return -1;
        }
        
        public SmoothReinforcement(double d, double b) { delta = d; beta = b; }
}
