package models;

import java.util.*;
import java.io.*;

import utilities.Game;
import utilities.GraphIO;
import utilities.Individual;
import utilities.Jobs;
import utilities.MersenneTwister;
import utilities.ModelError;

import learningRules.BgLearning;
import learningRules.LearningRule;
import learningRules.SmoothBGLearning;
import learningRules.StratReinforce;

import com.martiansoftware.jsap.*;

import connectionRules.ConnectNothing;
import connectionRules.ConnectionRule;

import mutationRules.MutateBeliefs;

/**
* This is a class that runs the BalaGoyal Model of Bayesian Learning in Networks.
* 
* @author Kevin Zollman
*
*/


public class BgModel {

        Game[] games;
        
        /**
        * Sets an array of games, one for each state of the world.
        * @param g The array of Game objects
        */
        public synchronized void setGames(Game[] g) { games = g;}
        
        /**
        * Returns the games for the BgModel
        * @return the games
        */
        public Game[] getGames() { return games; }
        
          // The folks in the network
        Individual[] individuals;
        /**
        * Sets the list of individuals for the model.
        * @param i An array of individuals
        */
        public synchronized void setIndividuals(Individual[] i) { individuals = i; }
        
        /**
        * Returns the set of individuals for the model.
        * @return the individuals
        */
        public Individual[] getIndividuals() {return individuals;}

        // The random number generator
        MersenneTwister random;
        /**
        * Sets the random number generator.
        * @param r the mersennetwister random number generator
        */
        public synchronized void setRandom(MersenneTwister r) { random =r; }
        
        /** 
        * Returns the random number generator for the model.
        * @return the mersenne twister random number generator
        */
        public MersenneTwister getRandom() { return random; }        
        
        
        /**
        * Forms a new network of specified size; doesn't expect much to be set.
        * This method only tells each individual of the network, does not set anything else.
        * @param inds A list of individuals
        */
        public synchronized void createPlayersLow(Individual[] inds) {
                individuals = inds;
                for(int i=0;i<inds.length;i++) {
                        individuals[i].setNetwork(individuals);
                }
        }
        /**
        * Forms a new network of specified size; expects random to be set.
        * This method tells each individual of the network, sets the random number generator,
        * and calle each individuals initStart function.
        *  
        * @param inds A list of individuals
        * @param s The number of states of the world
        * @throws ModelError If an initialization error
        */
        public synchronized void createPlayersHigh(Individual[] inds, int s) throws ModelError {
                individuals = inds;
                for(int i=0;i<inds.length;i++) {
                        individuals[i].setNetwork(individuals);
                        individuals[i].setRandom(random);
                        individuals[i].initStart(s);
                        
                }
        }
        
        /**
        * Forms a new network of specified size; expects random to be set.
        * This method tells each individual of the network, sets the random number generator,
        * gives the individuals a connection and learning rule, sets the mutation rate, sets 
        * belief constraint, and calls each individuals initStart function
        *
        * @param inds A list of individuals
        * @param s The number of states of the world
        * @param c The connection rule used for forming network connections by each individual
        * @param l The learning rule used for determining strategies by each individual
        * @param m The mutation rate for each individual to reset their beliefs
        * @param z The zero-state constraint for beliefs (initial beliefs cannot be lower than z)
        * @throws ModelError If something fails during initialization
        */
        public synchronized void createPlayersHigh(Individual[] inds, int s, ConnectionRule c, LearningRule l, double m, double z) throws ModelError{
                individuals = inds;
                for(int i=0;i<inds.length;i++) {
                        individuals[i].setNetwork(individuals);
                        individuals[i].setRandom(random);
                        individuals[i].setConnectionRule(c);
                        individuals[i].setLearningRule(l);
                        individuals[i].setMutationRule(new MutateBeliefs(m));
                        individuals[i].setConstraint(z);
                        individuals[i].initStart(s);

                }
        }
        
        /**
        * Forms a new network of specified size; expects random to be set.
        * This method tells each individual of the network, sets the random number generator,
        * gives the individuals a connection and learning rule, sets the list of games sets the 
        * mutation rate, sets belief constraint, and calls each individuals initStart function
        *
        * @param inds A list of individuals
        * @param s The number of states of the world
        * @param c The connection rule used for forming network connections by each individual
        * @param l The learning rule used for determining strategies by each individual
        * @param g The games used in the model
        * @param m The mutation rate for each individual to reset their beliefs
        * @param z The zero-state constraint for beliefs (initial beliefs cannot be lower than z)
        * @throws ModelError If something fails during initialization
        */
        public synchronized void createPlayersHigh(Individual[] inds, int s, ConnectionRule c, LearningRule l, Game g, double m, double z) throws ModelError{
                individuals = inds;
                for(int i=0;i<inds.length;i++) {
                        individuals[i].setNetwork(individuals);
                        individuals[i].setRandom(random);
                        individuals[i].setConnectionRule(c);
                        individuals[i].setLearningRule(l);
                        individuals[i].setGame(g);
                        individuals[i].setMutation(m);
                        individuals[i].setConstraint(z);
                        individuals[i].initStart(s);
                }
        }
        
        
        /**
        * Resets then plays a generation.
        * This first calls each individuals reset function, then calls each individuals playGame
        * function, then calles each individuals update function
        * @throws ModelError If something bad happens
        */
        public synchronized void playGen() throws ModelError {
                for(int i=0;i<individuals.length;i++) 
                        individuals[i].reset();
                for(int i=0;i<individuals.length;i++) 
                        individuals[i].playGame();
                for(int i=0;i<individuals.length;i++)
                        individuals[i].update();
        }
        
        // Returns three values, 0=unconverged, 1=converged to the true state, 2=converged to the
        //   wrong state.
        
        /**
        * Checks to see if the population has converged either to the correct state or 
        * the bad state.
        * This function presumes that state 0 is the uninformative state.
        * @param sow What is the actual state of the world?
        * @return This function returns 0 if unconverged, 1 if converged to the true state, 
        *         2 if converged to the wrong state (only possible in good states; i.e. not 
        *         state 0).
        */
        
        public synchronized int detectConvergence(int sow) throws ModelError {
                boolean trueconv = true;
                if(sow == 0) {
                        for(int k=0;k<getIndividuals().length;k++) {
                                if(getIndividuals()[k].getStrat() != 0) {
                                        trueconv = false;
                                        break; 
                                }
                        }
                        if(trueconv == true)
                                return 1;
                        else
                                return 0;
                }
                else {
                        boolean badconv = true;
                        for(int k=0;k<getIndividuals().length;k++) {
                                if(getIndividuals()[k].getBelief(sow) < 0.9999) 
                                        trueconv = false;
                                if(getIndividuals()[k].getStrat() != 0)
                                        badconv = false;
                                if(trueconv == false && badconv == false)
                                        break;
                                                                                
                        }
                        if(badconv == true) 
                                return 2;
                        if(trueconv == true)
                                return 1;
                        else
                                return 0;
                }
        }
        
        
        /**
        * Does nothing
        */
        public BgModel() {};
        
        /**
        * Creates a new BgModel with a specified number of individuals.
        * This creates a new list of individuals, sets their id number, and tells each about the 
        * network.
        * @param n The size of the network
        */
        public BgModel(int n) {
                individuals = new Individual[n];
                for(int i=0;i<n;i++) {
                        individuals[i] = new Individual();
                        individuals[i].setId(n);
                        individuals[i].setNetwork(individuals);
                }
        }
        
        
        /** 
        * Main, runs a model.
        * @param args The command line arguments
        */
        public static void main(String[] args) {
                
                
                try {

                // We're using the JSAP system for argument parsing
                
                JSAP jsap = new JSAP();
                
                // -j <jobfile> is the job-file
                FlaggedOption jobfile = new FlaggedOption("jobFile")
                                                .setShortFlag('j')
                                                .setLongFlag("jobfile")
                                                .setRequired(true)
                                                .setStringParser(JSAP.STRING_PARSER);
                jobfile.setHelp("The job file containing network information");
                jsap.registerParameter(jobfile);
                
                // -n does the runs with no bad state
                Switch switchnb = new Switch("noBadState")
                                        .setShortFlag('n')
                                        .setLongFlag("nobadstate");
                switchnb.setHelp("Only runs the good state");
                jsap.registerParameter(switchnb);
                                
                // -v prints detailed output
                FlaggedOption ver = new FlaggedOption("verbose")
                                                .setShortFlag('v')
                                                .setLongFlag("verbose")
                                                .setRequired(true)
                                                .setDefault("0")
                                                .setStringParser(JSAP.INTEGER_PARSER);
                ver.setHelp("0 only summary data, 1 is data on each run, 2 is data on every generation, 3 is beliefs on every generation");
                jsap.registerParameter(ver);
                
                // -r <runs> s the number of runs per job, default 40,000
                FlaggedOption runsoption = new FlaggedOption("runs")
                                                .setShortFlag('r')
                                                .setLongFlag("runs")
                                                .setRequired(true)
                                                .setDefault("40000")
                                                .setStringParser(JSAP.LONG_PARSER);
                runsoption.setHelp("Number of runs per job");
                jsap.registerParameter(runsoption);
                
                // -u <payoff> for the uninformative act, default 50
                FlaggedOption uninform = new FlaggedOption("uninformPayoff")
                                                .setShortFlag('u')
                                                .setLongFlag("uninformpayoff")
                                                .setRequired(true)
                                                .setDefault("50")
                                                .setStringParser(JSAP.INTEGER_PARSER);
                uninform.setHelp("Set the payoff for the uniformative act");
                jsap.registerParameter(uninform);
                
                // -g <generations> is the maximum number of generations until give up
                //                  default 2000
                FlaggedOption generations = new FlaggedOption("generations")
                                                .setShortFlag('g')
                                                .setLongFlag("generations")
                                                .setRequired(true)
                                                .setDefault("2000")
                                                .setStringParser(JSAP.INTEGER_PARSER);
                generations.setHelp("Number of generations per run");
                jsap.registerParameter(generations);
                
                
                // -s <stratnum> gives the number of strategies default 2
                FlaggedOption stratnum = new FlaggedOption("stratnum")
                                                .setShortFlag('s')
                                                .setLongFlag("stratnum")
                                                .setRequired(true)
                                                .setDefault("2")
                                                .setStringParser(JSAP.INTEGER_PARSER);
                stratnum.setHelp("Number of states/strategies");
                jsap.registerParameter(stratnum);
                
                // -b  sets the priors to be biased toward the bad state
                Switch biasedpriors = new Switch("biasedPriors")
                                        .setShortFlag('b')
                                        .setLongFlag("biasedpriors");
                biasedpriors.setHelp("Starts the priors biased toward the bad state");
                jsap.registerParameter(biasedpriors);
                
                // -m <rate> sets tbe mutation rate, default 0.0
                FlaggedOption mutationrate = new FlaggedOption("mutationRate")
                                                .setShortFlag('m')
                                                .setLongFlag("mutation")
                                                .setRequired(true)
                                                .setDefault("0.0")
                                                .setStringParser(JSAP.DOUBLE_PARSER);
                mutationrate.setHelp("Sets the probability that an individual randomly changes her beleifs");
                jsap.registerParameter(mutationrate);
                
                // -d turns of bad state convergence detection
                Switch switchbadconv = new Switch("noConvergenceDetect")
                                        .setShortFlag('d')
                                        .setLongFlag("noconvergencedetect");
                switchbadconv.setHelp("Turns off convergence detection for the bad state");
                jsap.registerParameter(switchbadconv);
                
                // -l learning rule
                FlaggedOption lr = new FlaggedOption("learningRule")
                                        .setShortFlag('l')
                                        .setLongFlag("learningrule")
                                        .setRequired(true)
                                        .setDefault("fp")
                                        .setStringParser(JSAP.STRING_PARSER);
                lr.setHelp("Sets the learning rule to use: fp for fictious play, sfp for smoothed ficitious play, re for reinforcement");
                jsap.registerParameter(lr);
                
                // -a gamma for smoothed fictitious play
                FlaggedOption gamma = new FlaggedOption("gamma")
                                                .setShortFlag('a')
                                                .setLongFlag("gamma")
                                                .setRequired(true)
                                                .setDefault("0.05")
                                                .setStringParser(JSAP.DOUBLE_PARSER);
                gamma.setHelp("Sets the gamma for smoothed fictitious play");
                jsap.registerParameter(gamma);
                
                
                // Process commandline 
                JSAPResult config = jsap.parse(args);   
                if (!config.success()) {
                        System.err.println();
                        System.err.println("Usage: java "
                                + BgModel.class.getName());
                        System.err.println("                "
                                + jsap.getUsage());
                        System.err.println();
                        System.err.println(jsap.getHelp());
                        System.exit(1);
                }
                if (config.getBoolean("biasedPriors") && ((config.getDouble("mutationRate") <= 0.0) || (!config.getBoolean("noConvergenceDetect")))) {
                        System.err.println();
                        System.err.println("When using biased priors, mutation rate must be higher than 0.0 and convergence detection must be turned off");
                        System.exit(1);
                }
                
                double zeroconstraint = 0.0;
                if(config.getBoolean("biasedPriors")) 
                        zeroconstraint = (double) (config.getInt("stratnum")-1)/config.getInt("stratnum");
                
                
                //    Jobs file
                HashSet jobs = GraphIO.readJobsFile(config.getString("jobFile"));
                Iterator todo = jobs.iterator();
                
                //Other initializions
                BgModel model = new BgModel();
                model.setRandom(new MersenneTwister());
                
                //Parameters of the model
                boolean stochastic = true;
                ConnectionRule c = new ConnectNothing();
                
                // The second argument gives the number of strategies, we will now generate a
                //   a game for that number of strategies.  If an agent assigns equal probability
                //   to all states, she is indifferent between the actions.
                
                int[][][] matrix = new int[config.getInt("stratnum")][config.getInt("stratnum")][];
                
                //Set up the bad payoff and good payoffs
                int badpayoff = config.getInt("uninformPayoff")-1;
                int goodpayoff = (config.getInt("uninformPayoff") * config.getInt("stratnum")) - (badpayoff*(config.getInt("stratnum")-1));
                int[] uninformmatrix = new int[config.getInt("stratnum")];
                int[] badmatrix = new int[config.getInt("stratnum")];
                int[] goodmatrix = new int[config.getInt("stratnum")];
                for(int i=0;i<config.getInt("stratnum");i++) {
                        uninformmatrix[i] = config.getInt("uninformPayoff");
                        badmatrix[i] = badpayoff;
                        goodmatrix[i] = goodpayoff;
                }
                
                //Set up the games
                Game[] games = new Game[config.getInt("stratnum")];
                for(int i=0;i<config.getInt("stratnum");i++) {
                        matrix[i][0] = uninformmatrix;
                        
                        for(int j=1;j<config.getInt("stratnum");j++) {
                                if(i==j) 
                                        matrix[i][j] = goodmatrix;
                                else
                                        matrix[i][j] = badmatrix;
                        }
                        games[i] = new Game(stochastic, model.getRandom());
                        games[i].setMatrix(matrix[i], badmatrix[0] + goodmatrix[0]);
                        games[i].setStrategic(false);
                }
                model.setGames(games);
                
                
                // Set the learning rule
                LearningRule l;
                if(config.getString("learningRule").equals("fp")) {
                        l = new BgLearning(model.getGames());
                }
                else {
                        if(config.getString("learningrule").equals("sfp")) {
                                l = new SmoothBGLearning(model.getGames(), config.getDouble("gamma"));
                        }
                        else {
                                l = new StratReinforce(0.01);
                        }
                }
                                
                System.out.println("#BalaGoyal Model parameter list: ");
                System.out.println("#   Jobfile: " + config.getString("jobFile"));
                System.out.println("#   NoBadState: " + config.getBoolean("noBadState"));
                System.out.println("#   Verbosity: " + config.getInt("verbose"));
                System.out.println("#   Runs: " + config.getLong("runs"));
                System.out.println("#   Number of Strategies/States: " + config.getInt("stratnum"));
                System.out.println("#   Uninformative Payoff: " + config.getInt("uninformPayoff"));
                System.out.println("#   Max. Generations/Run: " + config.getInt("generations"));
                System.out.println("#   Start with biased priors: " + config.getBoolean("biasedPriors"));
                System.out.println("#   Mutation Rate: " + config.getDouble("mutationRate"));
                System.out.println("#   Detect Bad Convergence: " + (!config.getBoolean("noConvergenceDetect")));
                
                while(todo.hasNext()) {
                
                Jobs job = (Jobs) todo.next();
                model.createPlayersHigh(GraphIO.adjacencyMatrix(job.getMatrix()), matrix.length, c, l, config.getDouble("mutationRate"), zeroconstraint);
                
                if(config.getInt("verbose")==0) {
                        System.out.println("----");
                        System.out.println("Job Id: " + job.getJobID());
                        System.out.println("Job Description: " + job.getDescrip());
                }
                
                long[] avegen = new long[config.getInt("stratnum")];
                long[] converged = new long[config.getInt("stratnum")];
                long badconverged = 0L;
                long badpriors = 0L;
                long[] statecount = new long[config.getInt("stratnum")];
                for(int i=0;i<config.getInt("stratnum");i++) 
                        avegen[i] = converged[i] = statecount[i] = 0L;
                for(long runs=0L;runs<config.getLong("runs");runs++) {
                        int sow;
                        if(config.getBoolean("noBadState")) {
                                sow = model.getRandom().nextInt(config.getInt("stratnum")-1) + 1;
                        }
                        else {
                                sow = model.getRandom().nextInt(config.getInt("stratnum"));
                        }
                        statecount[sow]++;
                        for(int j=0;j<model.getIndividuals().length;j++) {
                                model.getIndividuals()[j].randBeliefs(config.getInt("stratnum"));
                                model.getIndividuals()[j].setGame(model.getGames()[sow]);
                        }
                        double averageact = 0.0;
                        long averageactpay[] = new long[config.getInt("stratnum")];
                        int averageactcount[] = new int[config.getInt("stratnum")];
                        String actpaystring = "";
                        int i;
                        int end=0;
                        for(i=0;i<config.getInt("generations");i++) {
                                model.playGen();
                                end = model.detectConvergence(sow);
                                //this is the very verbose output
                                if(config.getInt("verbose") == 2) {
                                        averageact =0.0;
                                        actpaystring = "";
                                        for(int j=0;j<config.getInt("stratnum");j++) {
                                                averageactpay[j] = 0;
                                                averageactcount[j] = 0;
                                        }
                                        for(int j=0;j<model.getIndividuals().length;j++) {
                                                averageact += (double) ((double)1/model.getIndividuals().length) * 
                                                        model.getIndividuals()[j].getStrat();
                                                averageactpay[model.getIndividuals()[j].getStrat()] += 
                                                        model.getIndividuals()[j].getPayoff();
                                                averageactcount[model.getIndividuals()[j].getStrat()]++;
                                        }
                                        for(int j=0;j<config.getInt("stratnum");j++) 
                                                actpaystring += ((double) averageactpay[j]/
                                                        ((double) averageactcount[j]*model.getGames()[sow].getHighestPayoff())) + " ";
                                        
                                        System.out.println(job.getJobID() + " " + job.getDescrip() + 
                                                " " + sow + " " +runs + " " + i + " " + averageact + " " +actpaystring);
                                }
                                if(config.getInt("verbose") == 3) {
                                	
                                	double[] beliefAve = new double[model.getIndividuals()[0].getBeliefs()[0].length];
                                	for(int k=0;k<beliefAve.length;k++) 
                                		beliefAve[k] = 0;                                	
                                    for(int j=0;j<model.getIndividuals().length;j++) {
                                    	for(int k=0;k<beliefAve.length;k++) {
                                    		beliefAve[k] += model.getIndividuals()[j].getBeliefs()[0][k];
                                    	}
                                    }
                                    String aves = "";
                                    
                                    for(int k=0;k<beliefAve.length;k++) {
                                    	aves += ((double) beliefAve[k]/model.getIndividuals().length) + " ";
                                    }
                                    
                                    System.out.println(runs + " " + i + " " + " " + aves);
                                    
                                	
                            	}
                                // zero means not converged
                                if(end == 0) 
                                        continue;
                                // one means converged to correct state
                                if(end == 1)
                                        break;
                                // two means converged to the bad state when the state was good
                                if((end == 2) && (!config.getBoolean("noConvergenceDetect")))
                                        break;
                        }
                        if(i != config.getInt("generations")-1) {
                                if(end == 2) {
                                        badconverged++;
                                        if(i==0)
                                                badpriors++;
                                }
                                else {
                                        converged[sow]++;
                                        avegen[sow] += i;
                                }
                        }
                        switch(config.getInt("verbose")) {
                                case 0:
                                        if(runs%250==0) 
                                                System.out.print("*");
                                        break;
                                case 1: 
                                        System.out.println(job.getJobID() + " " + job.getDescrip() + 
                                                " " + runs + " " + sow + " " + end + " " + i);
                                        break;
                        }               
                }
                if(config.getInt("verbose")==0) {
                        System.out.println();
                        if(!config.getBoolean("noBadState")) {
                                System.out.println("State 0: Occured " + statecount[0] + " times, of these " +
                                        converged[0] + " runs converged to the correct ");
                                System.out.println("   beliefs in an average of " + 
                                        (double) avegen[0]/converged[0] + " generations.  ");
                        }
                        
                        long goodcount = 0L;
                        long goodconv = 0L;
                        long goodgen = 0L;
                        for(int i=1;i<config.getInt("stratnum");i++) {
                                goodcount += statecount[i];
                                goodgen += avegen[i];
                                goodconv += converged[i];
                        }
                        System.out.println("A good state occured " + (goodcount) +
                                " times, of these " + goodconv + 
                                " runs converged to the");
                        System.out.println("  correct action in an average of " +
                                (double) goodgen/goodconv + " generations.  ");
                                System.out.println("   " + badconverged + " converged to the wrong act (" + badpriors + " b/c of priors)"); 
                }
                }        
        }
        catch(ModelError e) {
                System.out.println(e);
        }
        catch(FileNotFoundException e) {
                System.out.println(e);
        }
        catch(IOException e) {
                System.out.println(e);
        }
        catch(JSAPException e) {
                System.out.println(e);
        }
        }
}
