package models;
import utilities.Game;
import utilities.Individual;
import utilities.MersenneTwister;
import utilities.ModelError;
import learningRules.LearningRule;
import learningRules.PayoffLearning;

import com.martiansoftware.jsap.*;

import connectionRules.ConnectNothing;

/**
* This class implements a learning in network model with biased network formation
*  This model uses beta-distribution learning (aka payoff learning).
* 
* @author Kevin Zollman
*
*/

public class BiasedNetworkModel {
        
        public static final MersenneTwister random = new MersenneTwister();
        
        Individual[] individs;	
        /**
	 * Returns the value of indivis.
	 */
	public Individual[] getIndivids() { return individs; }
	
        /**
	 * Sets the value of indivis.
	 * @param indivis The value to assign indivis.
	 */
        public void setIndivids(Individual[] individs) {this.individs = individs;}
        
        
        Game game;	
        /**
	 * Returns the value of game.
	 */
	public Game getGame(){ 	return game;	}
        
	/**
	 * Sets the value of game.
	 * @param game The value to assign game.
	 */
	public void setGame(Game game)	{this.game = game;}
        
        double minimumProb;
        
        /**
        * Gets the connection probability
        * @return The connection probability
        */
        public double getMinimumProb() { return minimumProb; }
        
        /**
        * Sets the connection probability
        * @param p The connection probability
        */
        public void setMinimumProb(double p) { minimumProb = p; }
       
         double maximumProb;
        
        /**
        * Gets the connection probability
        * @return The connection probability
        */
        public double getMaximumProb() { return maximumProb; }
        
        /**
        * Sets the connection probability
        * @param p The connection probability
        */
        public void setMaximumProb(double p) { maximumProb = p; }
        
        int totalPlayed;
	/**
	 * Returns the value of totalPlayed.
	 */
	public int getTotalPlayed()
	{
		return totalPlayed;
	}

	/**
	 * Sets the value of totalPlayed.
	 * @param totalPlayed The value to assign totalPlayed.
	 */
	public void setTotalPlayed(int totalPlayed)
	{
		this.totalPlayed = totalPlayed;
	}
   
        /**
        * Increments total count
        */
        public void incTotalPlayed() { totalPlayed++; }
        
        // I initially had each individual keep track of her own strategy history
        // that seemed to be taking far too long.  This is a second try for keeping track
        // of each individual pairwise agreement.  More memory, but faster computations
        
        //agreement i, j is only defined for i <= j
        int[][] agreement;
        
        /**
        * Resets the value for agreement.
        * @param n Number of agents in the network
        */
        public void resetAgreement(int n) {
                agreement = new int[n][n];
                for(int i=0;i<n;i++) {
                        for(int j=0;j<n;j++) {
                                agreement[i][j] = 0;
                        }
                }
        }
        
        /**
        * Returns the agreement count between i and j.  Although agreement is only defined for
        * i <= j, arguments can be passed in any order.  Since agreement is symmetric, 
        * getAgreement(i, j) == getAgreement(j, i) always.
        * @param i One agent
        * @param j Another
        */
        public int getAgreement(int i, int j) { return agreement[Math.min(i,j)][Math.max(i,j)]; }
        
        /**
        * Increments the agreement counter between i and j.  Since agreement is symmteric 
        * incAgreement(i,j) is equivalent to incAgreement(j, i)
        * @param i One agent
        * @param j Another
        */
        public void incAgreement(int i, int j) { agreement[Math.min(i,j)][Math.max(i,j)]++; }

        /**
        * Sets the agreement counter for two agents to a specified value.  Since
        * agreement is symmetric setAgreement(i,j,s) is equivalent to setAgreement(j,i,s)
        * @param i One agent
        * @param j Another
        * @param s The value to set
        */
        public void setAgreement(int i, int j, int a) { 
                agreement[Math.min(i,j)][Math.max(i,j)] = a;
        }
                
        
        /**
        * Resets then plays a generation.
        * This first calls each individuals reset function, then calls each individuals playGame
        * function, then calles each individuals update function
        * @throws ModelError If something bad happens
        */
        public synchronized void playGen() throws ModelError {
                
                // Modify the network for the next round
                for(int i=0;i<individs.length;i++) {
                        for(int j=i+1;j<individs.length;j++) {
                                double connectProb;
                                if(getTotalPlayed() == 0) {
                                        connectProb = minimumProb;
                                }
                                else {
                                        connectProb = (double) ((double) ((double) maximumProb - minimumProb)
                                                *((double) getAgreement(i,j)/getTotalPlayed()))
                                                + minimumProb;
                                }
                                System.out.print("M:" +  minimumProb + " P:" +connectProb + " A:" + getAgreement(i,j));
                                if(random.nextDouble() <= connectProb) {
                                        individs[i].addConnected(individs[j]);
                                        individs[j].addConnected(individs[i]);
                                }
                        }
                }


                // Standard game play
                
                // Resets for the new generation
                for(int i=0;i<individs.length;i++) 
                        individs[i].reset();
                // Plays the game
                for(int i=0;i<individs.length;i++) 
                        individs[i].playGame();
                // Calculates the agreement factors
                for(int i=0;i<individs.length;i++) {
                        for (int j=i+1;j<individs.length;j++) {
                                if(individs[i].getStrat() == individs[j].getStrat()) 
                                        incAgreement(i,j);
                        }
                }                        
                incTotalPlayed();
                // Updates based on the results, and clears the connections
                for(int i=0;i<individs.length;i++) {
                        individs[i].update();
                        individs[i].clearConnected();
                }
        }
        
        /**
        * Checks to see if the population has converged either to the correct state or 
        *   the bad state.  Convergence here only means individuals all persue the same
        *   strategy, so should not be used to stop a run.
        * @param optimal The optimal action
        * @return This function returns 0 if unconverged, 1 if converged to the true state, 
        *         2 if converged to the wrong state 
        */
        public synchronized int detectConvergence(int optimal) throws ModelError {
                int[] strats = new int[game.getStrats()];
                for(int k=0;k<game.getStrats();k++) 
                        strats[k] = 0;
                for(int k=0;k<getIndivids().length;k++) 
                        strats[getIndivids()[k].getStrat()]++;
                for(int k=0;k<game.getStrats();k++) {
                        if(strats[k] > 0) {
                                if(strats[k] < getIndivids().length)
                                        return 0;
                                else {
                                        if(k==optimal)
                                                return 1;
                                        else
                                                return 2;
                                }
                        }
                }
                return -1;
        }

        public BiasedNetworkModel(Individual[] i, Game g, LearningRule l, double m, 
                double min, double max) 
                throws ModelError {
                double[] zc = new double[g.getStrats()];
                for(int j=0;j<g.getStrats();j++)
                        zc[j] = 0.0;
                for(int j=0;j<i.length;j++) {
                        i[j].setNetwork(i);
                        i[j].setRandom(PayoffLearningModel.random);
                        i[j].setConnectionRule(new ConnectNothing());
                        i[j].setLearningRule(l);
                        i[j].setGame(g);
                        i[j].setMutation(m);
                        i[j].setConstraint(zc);
                }
                individs = i;
                game = g;
                minimumProb = min;
                maximumProb = max;
        }
        
        public static void main(String[] args) {
                try{
                
                        
                JSAP jsap = new JSAP();
                
                // -v prints detailed output
                FlaggedOption ver = new FlaggedOption("verbose")
                                                .setShortFlag('v')
                                                .setLongFlag("verbose")
                                                .setRequired(true)
                                                .setDefault("0")
                                                .setStringParser(JSAP.INTEGER_PARSER);
                ver.setHelp("0 only summary data, 1 is data on each run, 2 is data on every generation");
                jsap.registerParameter(ver);
                
                // -r <runs> s the number of runs per job, default 40,000
                FlaggedOption runsoption = new FlaggedOption("runs")
                                                .setShortFlag('r')
                                                .setLongFlag("runs")
                                                .setRequired(true)
                                                .setDefault("40000")
                                                .setStringParser(JSAP.LONG_PARSER);
                runsoption.setHelp("Number of runs per job");
                jsap.registerParameter(runsoption);
                
                // -b <max_payoff> is the maximum mean payoff for randomly assigning payoffs,
                //   default is 50
                FlaggedOption bigoption = new FlaggedOption("bigpayoff")
                                        .setShortFlag('b')
                                        .setLongFlag("bigpayoff")
                                        .setRequired(true)
                                        .setDefault("50")
                                        .setStringParser(JSAP.INTEGER_PARSER);
                bigoption.setHelp("Maxmimum mean payoff for randomly assigning payoffs");
                jsap.registerParameter(bigoption);
                
                // -g <generations> is the maximum number of generations until give up
                //                  default 2000
                FlaggedOption generations = new FlaggedOption("generations")
                                                .setShortFlag('g')
                                                .setLongFlag("generations")
                                                .setRequired(true)
                                                .setDefault("2000")
                                                .setStringParser(JSAP.INTEGER_PARSER);
                generations.setHelp("Number of generations per run");
                jsap.registerParameter(generations);
                
                // -s <stratnum> gives the number of strategies default 2
                FlaggedOption stratnum = new FlaggedOption("stratnum")
                                                .setShortFlag('s')
                                                .setLongFlag("stratnum")
                                                .setRequired(true)
                                                .setDefault("2")
                                                .setStringParser(JSAP.INTEGER_PARSER);
                stratnum.setHelp("Number of states/strategies");
                jsap.registerParameter(stratnum);
                
                // -m <rate> sets tbe mutation rate, default 0.0
                FlaggedOption mutationrate = new FlaggedOption("mutationRate")
                                                .setShortFlag('m')
                                                .setLongFlag("mutation")
                                                .setRequired(true)
                                                .setDefault("0.0")
                                                .setStringParser(JSAP.DOUBLE_PARSER);
                mutationrate.setHelp("Sets the probability that an individual randomly changes her beleifs");
                jsap.registerParameter(mutationrate);
                
                //A payoff to use
                FlaggedOption pay = new FlaggedOption("payoffs")
                                .setShortFlag('p')
                                .setLongFlag("payoff")
                                .setStringParser(JSAP.INTEGER_PARSER)
                                .setRequired(false)
                                .setAllowMultipleDeclarations(true);
                jsap.registerParameter(pay);
                
                // The maximum value for initial alphas and betas
                FlaggedOption ab = new FlaggedOption("alphabeta")
                                        .setShortFlag('a')
                                        .setLongFlag("alphabeta")
                                        .setStringParser(JSAP.INTEGER_PARSER)
                                        .setRequired(true)
                                        .setDefault("4");
                ab.setHelp("The maximum initial alpha and beta");
                jsap.registerParameter(ab);
                
                // -n The maximum n to go up to
                FlaggedOption num = new FlaggedOption("number")
                                                .setShortFlag('n')
                                                .setLongFlag("number")
                                                .setStringParser(JSAP.INTEGER_PARSER)
                                                .setRequired(true)
                                                .setDefault("6");
                num.setHelp("The maximum size up to which to run");
                jsap.registerParameter(num);
                
                                // Process commandline 
                JSAPResult config = jsap.parse(args);   
                if (!config.success()) {
                        System.err.println();
                        System.err.println("Usage: java "
                                + BgModel.class.getName());
                        System.err.println("                "
                                + jsap.getUsage());
                        System.err.println();
                        System.err.println(jsap.getHelp());
                        System.exit(1);
                }
                
                if(config.getInt("verbose") > 0) {
                        System.err.println("Sorry verbosity levels above 0 are not currently supported");
                        System.exit(1);
                }

                int[][] payoffs;
                int optimal=0;
                if(config.getInt("stratnum") != config.getIntArray("payoffs").length) {
                        payoffs = new int[config.getInt("stratnum")][config.getInt("stratnum")];
                        for(int i=0;i<config.getInt("stratnum");i++) {
                                if(i<config.getIntArray("payoffs").length) {
                                        for(int j=0;j<config.getInt("stratnum");j++)
                                                payoffs[i][j] = config.getIntArray("payoffs")[i];
                                        if(payoffs[i][0] > config.getInt("bigpayoff")) {
                                                System.err.println();
                                                System.err.println("Payoff cannot be larger than big payoff");
                                                System.exit(1);
                                        }
                                }
                                else {
                                        int number = PayoffLearningModel.random.nextInt(config.getInt("bigpayoff"));
                                        for(int j=0;j<config.getInt("stratnum");j++) 
                                                payoffs[i][j] = number;
                                }
                                if((payoffs[i][0] == payoffs[optimal][0]) && (i!=0)) {
                                        System.err.println();
                                        System.err.println("Cannot have two payoffs that are identical");
                                        System.exit(1);
                                }
                                if(payoffs[i][0] > payoffs[optimal][0])
                                        optimal = i;
                                
                        }
                        
                }
                else {
                        payoffs = new int[config.getInt("stratnum")][config.getInt("stratnum")];
                        for(int i=0;i<config.getInt("stratnum");i++) {
                                for(int j=0;j<config.getInt("stratnum");j++) 
                                        payoffs[i][j] = config.getIntArray("payoffs")[i];
                                
                                if((payoffs[i][0] == payoffs[optimal][0]) && (i!=0)) {
                                        System.err.println();
                                        System.err.println("Cannot have two payoffs that are identical: " 
                                                + i + " " + optimal);
                                        System.exit(1);
                                }
                                if(payoffs[i][0] > payoffs[optimal][0])
                                        optimal = i;
                                
                        }
                }
                
                //creates a new game with payoffs, it is stochastic, the rng, and its not strategic
                Game game = new Game(true, PayoffLearningModel.random);
                game.setMatrix(payoffs, config.getInt("bigpayoff"));
                game.setStrategic(false);
               
                LearningRule learning = new PayoffLearning(game);
                
                //print out parameters
                System.out.println("RandomNetworkModel parameter list:");
                System.out.println("    Verbosity: " + config.getInt("verbose"));
                System.out.println("    Runs: " + config.getLong("runs"));
                System.out.println("    Big Payoff: " + config.getInt("bigpayoff"));
                System.out.println("    Generations: " + config.getInt("generations"));
                System.out.println("    Strategies: " + config.getInt("stratnum"));
                System.out.println("    Mutation Rate: " + config.getDouble("mutationRate"));
                System.out.println("    Max alpha/beta: " + config.getInt("alphabeta"));
                System.out.println("    Max N: " + config.getInt("number"));
                for(int i=0;i<payoffs.length;i++)
                        System.out.println("    Payoff for strat " + i + ": " + payoffs[i][0]);
                
                // Do runs for each size from 2 to n
                for(int i=3;i<config.getInt("number");i++) {
                        
                                                                
                        
                        // Do runs for every probability between 1/(2 choose n) and 1
                        for(int min=1;min<=((i*(i-1))/2);min++) {
                        for(int max=min;max<=((i*(i-1))/2);max++) {
                                                              
                                Individual[] inds = new Individual[i];
                                for(int j=0;j<i;j++) 
                                        inds[j] = new Individual(j);
                        
                                BiasedNetworkModel model = new BiasedNetworkModel(inds, 
                                                        game, learning,
                                                        config.getDouble("mutationRate"),
                                                        ((double) min/((i*(i-1))/2)),
                                                        ((double) max/((i*(i-1))/2)));
                                
                                long[] convergence = new long[] {0L, 0L, 0L};
                                for(long k=0L;k<config.getLong("runs");k++) {
                                        model.setTotalPlayed(0);
                                        model.resetAgreement(model.getIndivids().length);
                                        for(int m=0;m<model.getIndivids().length;m++) { 
                          
                                               model.getIndivids()[m].randomAB(config.getInt("alphabeta"), config.getInt("alphabeta"), config.getInt("stratnum"));
                                                //model.getIndivids()[m].resetStratHist(config.getInt("generations"));
                                        }
                                        
                                        //System.out.print("*");
                                        int m=0;
                                        for(m=0;m<config.getInt("generations");m++) { 
                                                model.playGen();
                                        }
                                        int end = model.detectConvergence(optimal);
                                        convergence[end]++;
                                }
                        
                                if(config.getInt("verbose") == 0) {
                                        System.out.println(i + " " + ((double) min/((i*(i-1))/2)) + " "
                                                + ((double) max/((i*(i-1))/2)) + " "
                                                + ((double) convergence[0]/config.getLong("runs")) + " " 
                                                + ((double) convergence[1]/config.getLong("runs")) + " " 
                                                + ((double) convergence[2]/config.getLong("runs")));
                                }
                        }
                        
                }
                }
                }
                catch(ModelError e) {
                        System.out.println(e);
                }
                catch(JSAPException e) {
                        System.out.println(e);
                }
        }

        
}
