package models;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.HashSet;
import java.util.Iterator;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPException;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Switch;

import learningRules.LearningRule;
import learningRules.LearnNothing;
import mutationRules.MutateBeliefs;
import utilities.Game;
import utilities.GraphIO;
import utilities.Individual;
import utilities.Jobs;
import utilities.MersenneTwister;
import utilities.ModelError;
import connectionRules.ConnectNothing;
import connectionRules.ConnectionRule;

/**
 * This model has each individual privately play a two-armed Bandit and
 * then calculates a networks consensus probability from that.  The 
 * Bandit is one-arm-known like the BgModel bandit.
 * @author Kevin Zollman
 *
 */
public class DegrootBandit {

	
    Game[] games;
    
    /**
    * Sets an array of games, one for each state of the world.
    * @param g The array of Game objects
    */
    public synchronized void setGames(Game[] g) { games = g;}
    
    /**
    * Returns the games for the BgModel
    * @return the games
    */
    public Game[] getGames() { return games; }
    
      // The folks in the network
    Individual[] individuals;
    /**
    * Sets the list of individuals for the model.
    * @param i An array of individuals
    */
    public synchronized void setIndividuals(Individual[] i) { individuals = i; }
    
    /**
    * Returns the set of individuals for the model.
    * @return the individuals
    */
    public Individual[] getIndividuals() {return individuals;}

    // The random number generator
    MersenneTwister random;
    /**
    * Sets the random number generator.
    * @param r the mersennetwister random number generator
    */
    public synchronized void setRandom(MersenneTwister r) { random =r; }
    
    /** 
    * Returns the random number generator for the model.
    * @return the mersenne twister random number generator
    */
    public MersenneTwister getRandom() { return random; }        
    
    
    /**
    * Forms a new network of specified size; doesn't expect much to be set.
    * This method only tells each individual of the network, does not set anything else.
    * @param inds A list of individuals
    */
    public synchronized void createPlayersLow(Individual[] inds) {
            individuals = inds;
            for(int i=0;i<inds.length;i++) {
                    individuals[i].setNetwork(individuals);
            }
    }
    /**
    * Forms a new network of specified size; expects random to be set.
    * This method tells each individual of the network, sets the random number generator,
    * and calle each individuals initStart function.
    *  
    * @param inds A list of individuals
    * @param s The number of states of the world
    * @throws ModelError If an initialization error
    */
    public synchronized void createPlayersHigh(Individual[] inds, int s) throws ModelError {
            individuals = inds;
            for(int i=0;i<inds.length;i++) {
                    individuals[i].setNetwork(individuals);
                    individuals[i].setRandom(random);
                    individuals[i].initStart(s);
                    
            }
    }
    
    /**
    * Forms a new network of specified size; expects random to be set.
    * This method tells each individual of the network, sets the random number generator,
    * gives the individuals a connection and learning rule, sets the mutation rate, sets 
    * belief constraint, and calls each individuals initStart function
    *
    * @param inds A list of individuals
    * @param s The number of states of the world
    * @param c The connection rule used for forming network connections by each individual
    * @param l The learning rule used for determining strategies by each individual
    * @param m The mutation rate for each individual to reset their beliefs
    * @param z The zero-state constraint for beliefs (initial beliefs cannot be lower than z)
    * @throws ModelError If something fails during initialization
    */
    public synchronized void createPlayersHigh(Individual[] inds, int s, ConnectionRule c, LearningRule l, double m, double z) throws ModelError{
            individuals = inds;
            for(int i=0;i<inds.length;i++) {
                    individuals[i].setNetwork(individuals);
                    individuals[i].setRandom(random);
                    individuals[i].setConnectionRule(c);
                    individuals[i].setLearningRule(l);
                    individuals[i].setMutationRule(new MutateBeliefs(m));
                    individuals[i].setConstraint(z);
                    individuals[i].initStart(s);

            }
    }
    
    /**
    * Forms a new network of specified size; expects random to be set.
    * This method tells each individual of the network, sets the random number generator,
    * gives the individuals a connection and learning rule, sets the list of games sets the 
    * mutation rate, sets belief constraint, and calls each individuals initStart function
    *
    * @param inds A list of individuals
    * @param s The number of states of the world
    * @param c The connection rule used for forming network connections by each individual
    * @param l The learning rule used for determining strategies by each individual
    * @param g The games used in the model
    * @param m The mutation rate for each individual to reset their beliefs
    * @param z The zero-state constraint for beliefs (initial beliefs cannot be lower than z)
    * @throws ModelError If something fails during initialization
    */
    public synchronized void createPlayersHigh(Individual[] inds, int s, ConnectionRule c, LearningRule l, Game g, double m, double z) throws ModelError{
            individuals = inds;
            for(int i=0;i<inds.length;i++) {
                    individuals[i].setNetwork(individuals);
                    individuals[i].setRandom(random);
                    individuals[i].setConnectionRule(c);
                    individuals[i].setLearningRule(l);
                    individuals[i].setGame(g);
                    individuals[i].setMutation(m);
                    individuals[i].setConstraint(z);
                    individuals[i].initStart(s);
            }
    }
    
    /**
     * Performs initial experiment where each individual pulls an arm and then updates her beliefs
     * based on the outcome of that arm (but not the outcome obtained by her neighbors)
     * @throws ModelError
     */
    public synchronized void privatePlay() throws ModelError {
    	for(int i=0;i<individuals.length;i++) {
    		//reset
    		individuals[i].reset();
    		
    		//determine the maximum strat
    		double expectation[] = new double[games[0].getStrats()];
            
            for(int j=0;j<games[0].getStrats();j++)
                    expectation[j] = 0.0;
            for(int j=0;j<games[0].getStrats();j++) {
                    for(int k=0;k<games.length;k++) 
                           expectation[j] += (double) individuals[i].getBelief(k)*games[k].getExpectedPayoff(j);
            }
            
            double highestexpectation = 0.0;
            int strat = 0;
            for(int j=0;j<expectation.length;j++) {
                    if (expectation[j] > highestexpectation) {
                            strat = j;
                            highestexpectation = expectation[j];
                    }
            }
            individuals[i].setStrat(strat);
    	}
    	//play the game
    	for(int i=0;i<individuals.length;i++) 
            individuals[i].playGame();
    	
    	//calculate posterior
    	for(int i=0;i<individuals.length;i++) {
    		double denom = 0.0;
            double[] newBeliefs;
            
            // first based on your payoff
            for(int k=0;k<games.length;k++) 
                    denom += (double) games[k].getProb(individuals[i].getStrat(), 
                    		individuals[i].getPayoff())*individuals[i].getBelief(k);
            
            newBeliefs = new double[games.length];
            for(int k=0;k<games.length;k++) {
                    newBeliefs[k] = (double) ( (double) games[k].getProb(individuals[i].getStrat(),
                    		individuals[i].getPayoff()) * individuals[i].getBelief(k))/denom;
                    if(Double.isNaN(newBeliefs[k]))
                            throw new ModelError("NaN Belief: " + 
                            		games[k].getProb(individuals[i].getStrat(), 
                            				individuals[i].getPayoff()) +"*" + 
                            				individuals[i].getBelief(k) + " / " + denom);
            }
            individuals[i].setBeliefs(newBeliefs);
    	}
    		
    }
    
    /**
     * Calculate the consensus weights for a given network.  It presumes that individuals are not
     * self connected (that individual i is not in i.getConnected());
     * @return the weights
     */
    public synchronized double[] calcWeights() {
    	double[] weights = new double[individuals.length];
    	int totalEdges = 0;
    	
    	for(int i=0;i<weights.length;i++) {
    		weights[i] = individuals[i].getConnected().size();
    		totalEdges += weights[i];
    	}
    	for(int i=0;i<weights.length;i++) 
    		weights[i] = (double) (weights[i] + 1)/(totalEdges + weights.length);
    	
    	return weights;
    }
    
    /**
     * Calculates the consensus probabilities.
     * @param w The weights
     * @return The consensus probabilities
     */
    public synchronized double[] calcConsensus(double[] w) {
    	double[] consensus = new double[games.length];
    	
    	for(int i=0;i<consensus.length;i++) {
    		consensus[i] = 0.0;
    		for(int j=0;j<individuals.length;j++)
    			consensus[i] += (double) individuals[j].getBelief(i) * w[j];
    	}
    	return consensus;
    }
    
    /**
     * Checks to see if the population has converged either to the correct state or 
     * the bad state.
     * This function presumes that state 0 is the uninformative state.
     * @param sow What is the actual state of the world?
     * @return This function returns 0 if unconverged, 1 if converged to the true state, 
     *         2 if converged to the wrong state (only possible in good states; i.e. not 
     *         state 0).
     */
     
     public synchronized int detectConvergence(int sow) throws ModelError {
             boolean trueconv = true;
             if(sow == 0) {
                     for(int k=0;k<getIndividuals().length;k++) {
                             if(getIndividuals()[k].getStrat() != 0) {
                                     trueconv = false;
                                     break; 
                             }
                     }
                     if(trueconv == true)
                             return 1;
                     else
                             return 0;
             }
             else {
                     boolean badconv = true;
                     for(int k=0;k<getIndividuals().length;k++) {
                             if(getIndividuals()[k].getBelief(sow) < 0.9999) 
                                     trueconv = false;
                             if(getIndividuals()[k].getStrat() != 0)
                                     badconv = false;
                             if(trueconv == false && badconv == false)
                                     break;
                                                                             
                     }
                     if(badconv == true) 
                             return 2;
                     if(trueconv == true)
                             return 1;
                     else
                             return 0;
             }
     }
    
    /**
     * Does nothing
     */
     public DegrootBandit() {};
     
     /**
     * Creates a new BgModel with a specified number of individuals.
     * This creates a new list of individuals, sets their id number, and tells each about the 
     * network.
     * @param n The size of the network
     */
     public DegrootBandit(int n) {
             individuals = new Individual[n];
             for(int i=0;i<n;i++) {
                     individuals[i] = new Individual();
                     individuals[i].setId(n);
                     individuals[i].setNetwork(individuals);
             }
     }
    
	/**
	 * @param args Command line arguments
	 */
	public static void main(String[] args) {
		try {
//			 We're using the JSAP system for argument parsing
            
            JSAP jsap = new JSAP();
            
            // -j <jobfile> is the job-file
            FlaggedOption jobfile = new FlaggedOption("jobFile")
                                            .setShortFlag('j')
                                            .setLongFlag("jobfile")
                                            .setRequired(true)
                                            .setStringParser(JSAP.STRING_PARSER);
            jobfile.setHelp("The job file containing network information");
            jsap.registerParameter(jobfile);
            
            // -n does the runs with no bad state
            Switch switchnb = new Switch("noBadState")
                                    .setShortFlag('n')
                                    .setLongFlag("nobadstate");
            switchnb.setHelp("Only runs the good state");
            jsap.registerParameter(switchnb);
                            
            // -v prints detailed output
            FlaggedOption ver = new FlaggedOption("verbose")
                                            .setShortFlag('v')
                                            .setLongFlag("verbose")
                                            .setRequired(true)
                                            .setDefault("0")
                                            .setStringParser(JSAP.INTEGER_PARSER);
            ver.setHelp("0 only summary data, 1 is data on each run, 2 is data on every generation");
            jsap.registerParameter(ver);
            
            // -r <runs> s the number of runs per job, default 40,000
            FlaggedOption runsoption = new FlaggedOption("runs")
                                            .setShortFlag('r')
                                            .setLongFlag("runs")
                                            .setRequired(true)
                                            .setDefault("40000")
                                            .setStringParser(JSAP.LONG_PARSER);
            runsoption.setHelp("Number of runs per job");
            jsap.registerParameter(runsoption);
            
            // -u <payoff> for the uninformative act, default 50
            FlaggedOption uninform = new FlaggedOption("uninformPayoff")
                                            .setShortFlag('u')
                                            .setLongFlag("uninformpayoff")
                                            .setRequired(true)
                                            .setDefault("50")
                                            .setStringParser(JSAP.INTEGER_PARSER);
            uninform.setHelp("Set the payoff for the uniformative act");
            jsap.registerParameter(uninform);
            
            // -s <stratnum> gives the number of strategies default 2
            FlaggedOption stratnum = new FlaggedOption("stratnum")
                                            .setShortFlag('s')
                                            .setLongFlag("stratnum")
                                            .setRequired(true)
                                            .setDefault("2")
                                            .setStringParser(JSAP.INTEGER_PARSER);
            stratnum.setHelp("Number of states/strategies");
            jsap.registerParameter(stratnum);
            
//          -g <generations> is the maximum number of generations until give up
            //                  default 2000
            FlaggedOption generations = new FlaggedOption("generations")
                                            .setShortFlag('g')
                                            .setLongFlag("generations")
                                            .setRequired(true)
                                            .setDefault("2000")
                                            .setStringParser(JSAP.INTEGER_PARSER);
            generations.setHelp("Number of generations per run");
            jsap.registerParameter(generations);
            
            // Process commandline 
            JSAPResult config = jsap.parse(args);   
            if (!config.success()) {
                    System.err.println();
                    System.err.println("Usage: java "
                            + BgModel.class.getName());
                    System.err.println("                "
                            + jsap.getUsage());
                    System.err.println();
                    System.err.println(jsap.getHelp());
                    System.exit(1);
            }
            
//          Jobs file
            HashSet jobs = GraphIO.readJobsFile(config.getString("jobFile"));
            Iterator todo = jobs.iterator();
            
            //Other initializions
            DegrootBandit model = new DegrootBandit();
            model.setRandom(new MersenneTwister());
            
            //Parameters of the model
            boolean stochastic = true;
            ConnectionRule c = new ConnectNothing();
            
            // The second argument gives the number of strategies, we will now generate a
            //   a game for that number of strategies.  If an agent assigns equal probability
            //   to all states, she is indifferent between the actions.
            
            int[][][] matrix = new int[config.getInt("stratnum")][config.getInt("stratnum")][];
            
            //Set up the bad payoff and good payoffs
            int badpayoff = config.getInt("uninformPayoff")-1;
            int goodpayoff = (config.getInt("uninformPayoff") * config.getInt("stratnum")) - (badpayoff*(config.getInt("stratnum")-1));
            int[] uninformmatrix = new int[config.getInt("stratnum")];
            int[] badmatrix = new int[config.getInt("stratnum")];
            int[] goodmatrix = new int[config.getInt("stratnum")];
            for(int i=0;i<config.getInt("stratnum");i++) {
                    uninformmatrix[i] = config.getInt("uninformPayoff");
                    badmatrix[i] = badpayoff;
                    goodmatrix[i] = goodpayoff;
            }
            
            //Set up the games
            Game[] games = new Game[config.getInt("stratnum")];
            for(int i=0;i<config.getInt("stratnum");i++) {
                    matrix[i][0] = uninformmatrix;
                    
                    for(int j=1;j<config.getInt("stratnum");j++) {
                            if(i==j) 
                                    matrix[i][j] = goodmatrix;
                            else
                                    matrix[i][j] = badmatrix;
                    }
                    games[i] = new Game(stochastic, model.getRandom());
                    games[i].setMatrix(matrix[i], badmatrix[0] + goodmatrix[0]);
                    games[i].setStrategic(false);
            }
            model.setGames(games);
            
            
            // Set the learning rule
            LearningRule l = new LearnNothing();
            
            System.out.println("# DegrootBandit Model parameter list: ");
            System.out.println("#   Jobfile: " + config.getString("jobFile"));
            System.out.println("#   NoBadState: " + config.getBoolean("noBadState"));
            System.out.println("#   Verbosity: " + config.getInt("verbose"));
            System.out.println("#   Runs: " + config.getLong("runs"));
            System.out.println("#   Number of Strategies/States: " + config.getInt("stratnum"));
            System.out.println("#   Uninformative Payoff: " + config.getInt("uninformPayoff"));
            
            while(todo.hasNext()) {
                
            	Jobs job = (Jobs) todo.next();
                model.createPlayersHigh(GraphIO.adjacencyMatrix(job.getMatrix()), matrix.length, c, l, 0.0, 0.0);
                
                double[] weights = model.calcWeights();
                
                for(long runs=0L;runs<config.getLong("runs");runs++) {
                    int sow;
                    if(config.getBoolean("noBadState")) {
                            sow = model.getRandom().nextInt(config.getInt("stratnum")-1) + 1;
                    }
                    else {
                            sow = model.getRandom().nextInt(config.getInt("stratnum"));
                    }
                    for(int j=0;j<model.getIndividuals().length;j++) {
                            model.getIndividuals()[j].randBeliefs(config.getInt("stratnum"));
                            model.getIndividuals()[j].setGame(model.getGames()[sow]);
                    }

                   
                    int gens = 0;
                    int end = 0;
    
                    for(gens=0;gens<config.getInt("generations");gens++) {
                    	model.privatePlay();
                    	double[] results = model.calcConsensus(weights);
                    	for(int i=0;i<model.getIndividuals().length;i++) {
                    		model.getIndividuals()[i].setBeliefs((double[]) results.clone());
                    	}
                    	
                        end = model.detectConvergence(sow);
                        // zero means not converged
                        if(end == 0) 
                                continue;
                        // one means converged to correct state
                        if(end == 1)
                                break;
                        // two means converged to the bad state when the state was good
                        if(end == 2)
                                break;
                    }
                    
                    double[] results = model.calcConsensus(weights);
                    
                    String beliefString = "";
                    for(int i=0;i<results.length;i++) {
                    	beliefString += results[i] + " ";
                    }
                    System.out.println(job.getJobID() + " " + job.getDescrip() + 
                            " " + sow + " " + runs + " " + end + " " + gens + " " + beliefString);
                }
            }
            
		}
		catch(ModelError e) {
            System.out.println(e);
		}
		catch(FileNotFoundException e) {
			System.out.println(e);
		}
		catch(IOException e) {
            System.out.println(e);
		}
		catch(JSAPException e) {
			System.out.println(e);
		}
	}

}
