package models;

import java.util.*;
import java.io.*;

import utilities.Game;
import utilities.GraphIO;
import utilities.Individual;
import utilities.Jobs;
import utilities.MersenneTwister;
import utilities.ModelError;

import learningRules.LearningRule;
import learningRules.PayoffLearning;

import com.martiansoftware.jsap.*;

import connectionRules.ConnectNothing;
import connectionRules.ConnectionRule;


public class PayoffLearningModel {
        
        public static final MersenneTwister random = new MersenneTwister();
        
        Individual[] individs;	
        /**
	 * Returns the value of indivis.
	 */
	public Individual[] getIndivids() { return individs; }

	/**
	 * Sets the value of indivis.
	 * @param indivis The value to assign indivis.
	 */
        public void setIndivids(Individual[] individs) {this.individs = individs;}
        
        Game game;	
        /**
	 * Returns the value of game.
	 */
	public Game getGame(){ 	return game;	}
        
	/**
	 * Sets the value of game.
	 * @param game The value to assign game.
	 */
	public void setGame(Game game)	{this.game = game;}
        
        /**
        * Resets then plays a generation.
        * This first calls each individuals reset function, then calls each individuals playGame
        * function, then calles each individuals update function
        * @throws ModelError If something bad happens
        */
        public synchronized void playGen() throws ModelError {
                for(int i=0;i<individs.length;i++) 
                        individs[i].reset();
                for(int i=0;i<individs.length;i++) 
                        individs[i].playGame();
                for(int i=0;i<individs.length;i++)
                        individs[i].update();
        }
        
        /**
        * Checks to see if the population has converged either to the correct state or 
        *   the bad state.  Convergence here only means individuals all persue the same
        *   strategy, so should not be used to stop a run.
        * @param optimal The optimal action
        * @return This function returns 0 if unconverged, 1 if converged to the true state, 
        *         2 if converged to the wrong state 
        */
        public synchronized int detectConvergence(int optimal) throws ModelError {
                int[] strats = new int[game.getStrats()];
                for(int k=0;k<game.getStrats();k++) 
                        strats[k] = 0;
                for(int k=0;k<getIndivids().length;k++) 
                        strats[getIndivids()[k].getStrat()]++;
                for(int k=0;k<game.getStrats();k++) {
                        if(strats[k] > 0) {
                                if(strats[k] < getIndivids().length)
                                        return 0;
                                else {
                                        if(k==optimal)
                                                return 1;
                                        else
                                                return 2;
                                }
                        }
                }
                return -1;
        }
        
        public PayoffLearningModel(Individual[] i, Game g, ConnectionRule c, LearningRule l, double m) 
                throws ModelError {
                double[] zc = new double[g.getStrats()];
                for(int j=0;j<g.getStrats();j++)
                        zc[j] = 0.0;
                for(int j=0;j<i.length;j++) {
                        i[j].setNetwork(i);
                        i[j].setRandom(PayoffLearningModel.random);
                        i[j].setConnectionRule(c);
                        i[j].setLearningRule(l);
                        i[j].setGame(g);
                        i[j].setMutation(m);
                        i[j].setConstraint(zc);
                }
                individs = i;
                game = g;
        }
        
        public static void main(String[] args) {
                try{
                
                        
                JSAP jsap = new JSAP();
                
                // -j <jobfile> is the job-file
                FlaggedOption jobfile = new FlaggedOption("jobFile")
                                                .setShortFlag('j')
                                                .setLongFlag("jobfile")
                                                .setRequired(true)
                                                .setStringParser(JSAP.STRING_PARSER);
                jobfile.setHelp("The job file containing network information");
                jsap.registerParameter(jobfile);

                // -v prints detailed output
                FlaggedOption ver = new FlaggedOption("verbose")
                                                .setShortFlag('v')
                                                .setLongFlag("verbose")
                                                .setRequired(true)
                                                .setDefault("0")
                                                .setStringParser(JSAP.INTEGER_PARSER);
                ver.setHelp("0 only summary data, 1 is data on each run, 2 is data on every generation");
                jsap.registerParameter(ver);
                
                // -r <runs> s the number of runs per job, default 40,000
                FlaggedOption runsoption = new FlaggedOption("runs")
                                                .setShortFlag('r')
                                                .setLongFlag("runs")
                                                .setRequired(true)
                                                .setDefault("40000")
                                                .setStringParser(JSAP.LONG_PARSER);
                runsoption.setHelp("Number of runs per job");
                jsap.registerParameter(runsoption);
                
                // -b <max_payoff> is the maximum mean payoff for randomly assigning payoffs,
                //   default is 50
                FlaggedOption bigoption = new FlaggedOption("bigpayoff")
                                        .setShortFlag('b')
                                        .setLongFlag("bigpayoff")
                                        .setRequired(true)
                                        .setDefault("50")
                                        .setStringParser(JSAP.INTEGER_PARSER);
                bigoption.setHelp("Maxmimum mean payoff for randomly assigning payoffs");
                jsap.registerParameter(bigoption);
                
                // -g <generations> is the maximum number of generations until give up
                //                  default 2000
                FlaggedOption generations = new FlaggedOption("generations")
                                                .setShortFlag('g')
                                                .setLongFlag("generations")
                                                .setRequired(true)
                                                .setDefault("2000")
                                                .setStringParser(JSAP.INTEGER_PARSER);
                generations.setHelp("Number of generations per run");
                jsap.registerParameter(generations);
                
                // -s <stratnum> gives the number of strategies default 2
                FlaggedOption stratnum = new FlaggedOption("stratnum")
                                                .setShortFlag('s')
                                                .setLongFlag("stratnum")
                                                .setRequired(true)
                                                .setDefault("2")
                                                .setStringParser(JSAP.INTEGER_PARSER);
                stratnum.setHelp("Number of states/strategies");
                jsap.registerParameter(stratnum);
                
                // -m <rate> sets tbe mutation rate, default 0.0
                FlaggedOption mutationrate = new FlaggedOption("mutationRate")
                                                .setShortFlag('m')
                                                .setLongFlag("mutation")
                                                .setRequired(true)
                                                .setDefault("0.0")
                                                .setStringParser(JSAP.DOUBLE_PARSER);
                mutationrate.setHelp("Sets the probability that an individual randomly changes her beleifs");
                jsap.registerParameter(mutationrate);
                
                //A payoff to use
                FlaggedOption pay = new FlaggedOption("payoffs")
                                .setShortFlag('p')
                                .setLongFlag("payoff")
                                .setStringParser(JSAP.INTEGER_PARSER)
                                .setRequired(false)
                                .setAllowMultipleDeclarations(true);
                jsap.registerParameter(pay);
                
                // The maximum value for initial alphas and betas
                FlaggedOption ab = new FlaggedOption("alphabeta")
                                        .setShortFlag('a')
                                        .setLongFlag("alphabeta")
                                        .setStringParser(JSAP.INTEGER_PARSER)
                                        .setRequired(true)
                                        .setDefault("4");
                ab.setHelp("The maximum initial alpha and beta");
                jsap.registerParameter(ab);
                
                
                // Process commandline 
                JSAPResult config = jsap.parse(args);   
                if (!config.success()) {
                        System.err.println();
                        System.err.println("Usage: java "
                                + BgModel.class.getName());
                        System.err.println("                "
                                + jsap.getUsage());
                        System.err.println();
                        System.err.println(jsap.getHelp());
                        System.exit(1);
                }
                
                int[][] payoffs;
                int optimal=0;
                if(config.getInt("stratnum") != config.getIntArray("payoffs").length) {
                        payoffs = new int[config.getInt("stratnum")][config.getInt("stratnum")];
                        for(int i=0;i<config.getInt("stratnum");i++) {
                                if(i<config.getIntArray("payoffs").length) {
                                        for(int j=0;j<config.getInt("stratnum");j++)
                                                payoffs[i][j] = config.getIntArray("payoffs")[i];
                                        if(payoffs[i][0] > config.getInt("bigpayoff")) {
                                                System.err.println();
                                                System.err.println("Payoff cannot be larger than big payoff");
                                                System.exit(1);
                                        }
                                }
                                else {
                                        int number = PayoffLearningModel.random.nextInt(config.getInt("bigpayoff"));
                                        for(int j=0;j<config.getInt("stratnum");j++) 
                                                payoffs[i][j] = number;
                                }
                                if((payoffs[i][0] == payoffs[optimal][0]) && (i!=0)) {
                                        System.err.println();
                                        System.err.println("Cannot have two payoffs that are identical");
                                        System.exit(1);
                                }
                                if(payoffs[i][0] > payoffs[optimal][0])
                                        optimal = i;
                                
                        }
                        
                }
                else {
                        payoffs = new int[config.getInt("stratnum")][config.getInt("stratnum")];
                        for(int i=0;i<config.getInt("stratnum");i++) {
                                for(int j=0;j<config.getInt("stratnum");j++) 
                                        payoffs[i][j] = config.getIntArray("payoffs")[i];
                                
                                if((payoffs[i][0] == payoffs[optimal][0]) && (i!=0)) {
                                        System.err.println();
                                        System.err.println("Cannot have two payoffs that are identical: " 
                                                + i + " " + optimal);
                                        System.exit(1);
                                }
                                if(payoffs[i][0] > payoffs[optimal][0])
                                        optimal = i;
                                
                        }
                }
                
                //creates a new game with payoffs, it is stochastic, the rng, and its not strategic
                Game game = new Game(true, PayoffLearningModel.random);
                game.setMatrix(payoffs, config.getInt("bigpayoff"));
                game.setStrategic(false);
               
                //Jobs file
                HashSet jobs = GraphIO.readJobsFile(config.getString("jobFile"));
                Iterator todo = jobs.iterator();
                
                ConnectionRule connect = new ConnectNothing();
                LearningRule learning = new PayoffLearning(game);
                
                //print out parameters
                System.out.println("PayoffLearningModel parameter list:");
                System.out.println("    JobFile: " + config.getString("jobFile"));
                System.out.println("    Verbosity: " + config.getInt("verbose"));
                System.out.println("    Runs: " + config.getLong("runs"));
                System.out.println("    Big Payoff: " + config.getInt("bigpayoff"));
                System.out.println("    Generations: " + config.getInt("generations"));
                System.out.println("    Strategies: " + config.getInt("stratnum"));
                System.out.println("    Mutation Rate: " + config.getDouble("mutationRate"));
                System.out.println("    Max alpha/beta: " + config.getInt("alphabeta"));
                for(int i=0;i<payoffs.length;i++)
                        System.out.println("    Payoff for strat " + i + ": " + payoffs[i][0]);
                
                while(todo.hasNext()) {
                        Jobs job = (Jobs) todo.next();
                        PayoffLearningModel model = new PayoffLearningModel(GraphIO.adjacencyMatrix(job.getMatrix()),
                                                                  game,
                                                                  connect,
                                                                  learning,
                                                                  config.getDouble("mutationRate"));
                        if(config.getInt("verbose")==0) {
                                System.out.println("----");
                                System.out.println("Job Id: " + job.getJobID());
                                System.out.println("Job Description: " + job.getDescrip());
                        }
                        
                        long[] convergence = new long[] {0L, 0L, 0L};
                        for(long i=0L;i<config.getLong("runs");i++) {
                                for(int j=0;j<model.getIndivids().length;j++) 
                                       model.getIndivids()[j].randomAB(config.getInt("alphabeta"), config.getInt("alphabeta"), config.getInt("stratnum"));
                                
                                int j=0;
                                for(j=0;j<config.getInt("generations");j++) { 
                                        model.playGen();
                                        if(config.getInt("verbose") ==2) {
                                                double[] averageactpay = new double[config.getInt("stratnum")];
                                                double[] averageactcount = new double[config.getInt("stratnum")];
                                                double averageact =0.0;
                                                String actpaystring = "";
                                                for(int k=0;k<config.getInt("stratnum");k++) {
                                                        averageactpay[k] = 0;
                                                        averageactcount[k] = 0;
                                                }
                                                for(int k=0;k<model.getIndivids().length;k++) {
                                                        averageact += (double) ((double)1/model.getIndivids().length) * 
                                                                model.getIndivids()[k].getStrat();
                                                        averageactpay[model.getIndivids()[k].getStrat()] += 
                                                                model.getIndivids()[k].getPayoff();
                                                        averageactcount[model.getIndivids()[k].getStrat()]++;
                                                }
                                                for(int k=0;k<config.getInt("stratnum");k++) 
                                                        actpaystring += ((double) averageactpay[k]/
                                                                ((double) averageactcount[k]*model.getGame().getHighestPayoff())) + " ";
                                        
                                                System.out.println(job.getJobID() + " " + job.getDescrip() + 
                                                        " " + optimal + " " + i + " " + j + " " + averageact + " " +actpaystring);
                                        }
                                }
                                int end = model.detectConvergence(optimal);
                                convergence[end]++;
                                switch(config.getInt("verbose")) {
                                case 0:
                                        if(i%250==0) 
                                                System.out.print("*");
                                        break;
                                case 1: 
                                        System.out.println(job.getJobID() + " " + job.getDescrip() + 
                                                " "  + optimal + " " + i + " "+ end + " " + j);
                                        break;
                                case 2:
                                        System.out.println("** New run **");
                                        break;
                                }
                        }
                        if(config.getInt("verbose")==0) {
                                System.out.println();
                                System.out.println("Out of " + config.getLong("runs") + " runs, " +
                                        convergence[0] + " did not converge, ");
                                System.out.println("    " + convergence[1] +
                                        " converged correctly, and ");
                                System.out.println("    " + convergence[2] + 
                                        " converged incorrectly");
                                System.out.println("     in a total of " + config.getInt("generations")
                                        + " generations.");
                        }

                }
                }
                catch(ModelError e) {
                        System.out.println(e);
                }
                catch(FileNotFoundException e) {
                        System.out.println(e);
                }
                catch(IOException e) {
                        System.out.println(e);
                }
                catch(JSAPException e) {
                        System.out.println(e);
                }
        }
}
