package models;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.HashSet;
import java.util.Iterator;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPException;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Switch;

import utilities.GraphIO;
import utilities.Individual;
import utilities.Jobs;
import utilities.MersenneTwister;
import utilities.ModelError;
import learningRules.CondorcetLearning;
import learningRules.LearningRule;
import mutationRules.MutateStrategy;
import mutationRules.MutationRule;
import connectionRules.ConnectionRule;
import connectionRules.ConnectNothing;


/**
* The world is set in state 0, each individual is given (noisy) access to the world.  With
* probability (1-e) they are given 0, with probability e they are given 1.  After their 
* initial assignments, each individual then polls her neighbors, and updates her belief based 
* on what the majority of her neighbors think.  (Each individual stays the same in case of tie.)
* This is equivalent to myopic best-response in a game of pure coordination.  In fact,
* this is how the model in instantiate here, although the results are translated to conform
* to the model described above.
* @author Kevin Zollman
*/
public class SimpleCondorcetModel {
	Individual[] individuals;
    
	/**
    * Sets the list of individuals for the model.
    * @param i An array of individuals
    */
    public synchronized void setIndividuals(Individual[] i) { individuals = i; }
    
    /**
    * Returns the set of individuals for the model.
    * @return the individuals
    */
    public Individual[] getIndividuals() {return individuals;}

    // The random number generator
    MersenneTwister random;
    
    /**
    * Sets the random number generator.
    * @param r the mersennetwister random number generator
    */
    public synchronized void setRandom(MersenneTwister r) { random =r; }
    
    /** 
    * Returns the random number generator for the model.
    * @return the mersenne twister random number generator
    */
    public MersenneTwister getRandom() { return random; }

    
    /**
     * This is used for enumerations, it takes a sequence
     * of numbers and a maximum for each value and increments
     * to the next in the sequence.  The first in the sequence
     * is 0,0,...,0 the last is n,n,...,n
     * @param s The sequence
     * @param n The maximum number
     * @returns The next in the sequence if defined, otherwise -1, -1, ..., -1
     */ 
    public int[] incrementSequence(int[] s, int n) {
    	int carry = 1;
    	for(int i=1;i<=s.length;i++) {
    		if(carry == 1) {
    			if(s[s.length-i] == (n-1)) {
    				s[s.length-i] = 0;
    			}
    			else {
    				s[s.length-i]++;
    				carry = 0;
    			}
    		}
    	}
    	if(carry == 1) {
    		int[] r = new int[s.length];
    		for(int i=0;i<s.length;i++) {
    			r[i] = -1;
    		}
    		return r;
    	}
    	return s;
    }
    
    /**
     * Sets up the individuals
     * @param is The network of individuals (already appropriately connected)
     * @param mr The mutation rule
     * @param lr The learning rule
     * @param cr The connection rule
     */
    public void createPlayers(Individual[] is, MutationRule mr, LearningRule lr, ConnectionRule cr ) {
    	individuals = is;
    	for(int i=0;i<is.length;i++) {
    		individuals[i].setNetwork(individuals);
    		individuals[i].setRandom(random);
    		individuals[i].setConnectionRule(cr);
    		individuals[i].setLearningRule(lr);
    		individuals[i].setMutationRule(mr);
    	}
    }

    /**
     * Resets the players beliefs
     * @param s The number of possible belief states
     * @param e The probability of believing belief state 0 (i.e. the truth)
     * @throws ModelError
     */
    public void resetPlayers(int s, double e) throws ModelError {
    	for(int i=0;i<individuals.length;i++) {
    		if(random.nextDouble() < ((double) 1 - e)) {
    			individuals[i].setStrat(0);
    			individuals[i].setLastStrat();
    		}
    		else {
    			individuals[i].setStrat(random.nextInt(s-1)+1);
    			individuals[i].setLastStrat();
    		}
    	}
    }
    
    /**
     * Detects if the individuals are in one of three states,
     *   unanimity on the truth (strategy 0), majority on the truth
     *   or none
     * @return 2 for unanimity, 1 for majority, and 0 for neither
     */
    public int detectState() throws ModelError {
    	int truecount = 0;
    	for(int i=0;i<individuals.length;i++) {
    		if(individuals[i].getStrat() == 0) 
    			truecount++;
    	}
    	if(truecount == individuals.length) 
    		return 2;
    	if(truecount > (individuals.length - truecount))
    		return 1;
    	return 0;
    }
    
    /**
     * Returns the count of individuals who believe the truth (strategy
     * 0).  
     * @return Number that believe the truth
     */
    public int countTruth() throws ModelError {
    	int truecount = 0;
    	for(int i=0;i<individuals.length;i++) {
    		if(individuals[i].getStrat() == 0) 
    			truecount++;
    	}
    	return truecount;
    }
    
    
    /**
     * Runs the model.  If there is mutation, it runs until the value g is obtained.
     * If there is no mutation, then it checks for convergence and stops if there is
     * convergence (or if g is obtained).
     * @param mr The mutation rate
     * @param tr A boolean saying if "truerate" is set (if true no convergence detection)
     * @param g The maximum number of generations
     * @param v Is the verbosity level (>2 means it prints something out for every gen)
     * @throws ModelError
     */
    public int runmodel(double mr, boolean tr, int g, int v) throws ModelError {
    	// If there is mutation, run until max gen is hit
		int i; 
    	
    	if((mr > 0.0) || tr) {
			 // runs the model
			 for(i=0;i<g;i++) {
				 if(v > 1) {
					 String output = new String(i + ":");
					 for(int j=0;j<individuals.length;j++) {
						 output += " " + individuals[j].getStrat();
					 }
					 
					 System.out.println(output);
				 }
				 for(int j=0;j<individuals.length;j++)
					 individuals[j].reset();
				 
			 }
		 }
		 // Otherwise, check for convergence
		 else {
			 int[] laststate = new int[individuals.length];
			 for(i=0;i<g;i++) {
				 if(v > 1) {
					 String output = new String(i + ":");
					 for(int j=0;j<individuals.length;j++) {
						 output += " " + individuals[j].getStrat();
					 }
					 
					 System.out.println(output);
				 }
				 
				 for(int j=0;j<individuals.length;j++) 
					 individuals[j].reset();
				 
				 boolean converged = true;
				 for(int j=0;j<individuals.length;j++) {
					 if(laststate[j] != individuals[j].getStrat()) {
						 converged = false;
						 break;
					 }
				 }
				 if(converged) break;

				 for(int j=0;j<individuals.length;j++) 
					 laststate[j] = individuals[j].getStrat();

			 }
		 }
    	if(v > 1) {
			 String output = new String(((int) i+1) + ":");
			 for(int j=0;j<individuals.length;j++) {
				 output += " " + individuals[j].getStrat();
			 }
			 
			 System.out.println(output);
		 }
		 return i;
    }
    
    /**
     * Does nothing
     */
    public SimpleCondorcetModel() {}
    
    /**
     * Sets up a group of individuals of size n.
     * @param n
     */
    public SimpleCondorcetModel(int n) {
    	individuals = new Individual[n];
        for(int i=0;i<n;i++) {
                individuals[i] = new Individual();
                individuals[i].setId(n);
                individuals[i].setNetwork(individuals);
        }
    }
 
    /** 
     * Main, runs a model.
     * @param args The command line arguments
     */
     public static void main(String[] args) {
             try{
            	 /*************** Command Line Processing *****************/
                 
                 JSAP jsap = new JSAP();
                 
                 // -j <jobfile> is the job-file
                 FlaggedOption jobfile = new FlaggedOption("jobFile")
                                                 .setShortFlag('j')
                                                 .setLongFlag("jobfile")
                                                 .setRequired(true)
                                                 .setStringParser(JSAP.STRING_PARSER);
                 jobfile.setHelp("The job file containing network information");
                 jsap.registerParameter(jobfile);
                                 
                 // -v prints detailed output
                 FlaggedOption ver = new FlaggedOption("verbose")
                                                 .setShortFlag('v')
                                                 .setLongFlag("verbose")
                                                 .setRequired(true)
                                                 .setDefault("0")
                                                 .setStringParser(JSAP.INTEGER_PARSER);
                 ver.setHelp("0 only summary data, 1 is data on each run, 2 is data on every generation");
                 jsap.registerParameter(ver);
                 
                 // -r <runs> s the number of runs per job, default 40,000
                 FlaggedOption runsoption = new FlaggedOption("runs")
                                                 .setShortFlag('r')
                                                 .setLongFlag("runs")
                                                 .setRequired(true)
                                                 .setDefault("40000")
                                                 .setStringParser(JSAP.LONG_PARSER);
                 runsoption.setHelp("Number of runs per job");
                 jsap.registerParameter(runsoption);
                 
                 // -g <generations> is the maximum number of generations until give up
                 //                  default 2000
                 FlaggedOption generations = new FlaggedOption("generations")
                                                 .setShortFlag('g')
                                                 .setLongFlag("generations")
                                                 .setRequired(true)
                                                 .setDefault("2000")
                                                 .setStringParser(JSAP.INTEGER_PARSER);
                 generations.setHelp("Number of generations per run");
                 jsap.registerParameter(generations);
                 
                 
                 // -s <stratnum> gives the number of strategies default 2
                 FlaggedOption stratnum = new FlaggedOption("stratnum")
                                                 .setShortFlag('s')
                                                 .setLongFlag("stratnum")
                                                 .setRequired(true)
                                                 .setDefault("2")
                                                 .setStringParser(JSAP.INTEGER_PARSER);
                 stratnum.setHelp("Number of states/strategies");
                 jsap.registerParameter(stratnum);
                 
                 // -m <rate> sets tbe mutation rate, default 0.0
                 FlaggedOption mutationrate = new FlaggedOption("mutationRate")
                                                 .setShortFlag('m')
                                                 .setLongFlag("mutation")
                                                 .setRequired(true)
                                                 .setDefault("0.0")
                                                 .setStringParser(JSAP.DOUBLE_PARSER);
                 mutationrate.setHelp("Sets the probability that an individual randomly changes her beleifs");
                 jsap.registerParameter(mutationrate);
                 
                 // -t <rate> sets the mutation rate from a true belief to a non-true one
                 FlaggedOption truerate = new FlaggedOption("truerate")
                 								.setShortFlag('t')
                 								.setLongFlag("truerate")
                 								.setRequired(false)
                 								.setStringParser(JSAP.DOUBLE_PARSER);
                 truerate.setHelp("Sets the mutation rate from a true belief to a non-true one (overrides -m)");
                 jsap.registerParameter(truerate);
                 
                 // -f <rate> sets the mutation rate from a non-true belief to any other
                 FlaggedOption falserate = new FlaggedOption("falserate")
                 								.setShortFlag('f')
                 								.setLongFlag("falserate")
                 								.setRequired(false)
                 								.setStringParser(JSAP.DOUBLE_PARSER);
                 falserate.setHelp("Sets the mutation rate from a false beleif to any other (overrides -m)");
                 jsap.registerParameter(falserate);
                 
                 // -e <number> is epsilon, the error rate
                 FlaggedOption epsilon = new FlaggedOption("epsilon")
                 							 .setShortFlag('e')
                 							 .setLongFlag("epsilon")
                 							 .setRequired(true)
                 							 .setDefault("0.1")
                 							 .setStringParser(JSAP.DOUBLE_PARSER);
                 epsilon.setHelp("Set epsilon, the error probability");
                 jsap.registerParameter(epsilon);
                 
                 // -c tells the system to enumerate all possibilities, not simulate
                 Switch enumerate = new Switch("enumerate")
                 						.setShortFlag('c')
                 						.setLongFlag("enumerate");
                 enumerate.setHelp("Tells the system to enumerate all possibilities instead of simulation");
                 jsap.registerParameter(enumerate);						
                 
                 // Process commandline 
                 JSAPResult config = jsap.parse(args);   
                 if (!config.success()) {
                         System.err.println();
                         System.err.println("Usage: java "
                                 + BgModel.class.getName());
                         System.err.println("                "
                                 + jsap.getUsage());
                         System.err.println();
                         System.err.println(jsap.getHelp());
                         System.exit(1);
                 }
                 
                 /********************* Set up *********************************/
                 // Jobs file
                 HashSet jobs = GraphIO.readJobsFile(config.getString("jobFile"));
                 Iterator todo = jobs.iterator();
                 
                 // Other initializions
                 SimpleCondorcetModel model = new SimpleCondorcetModel();
                 model.setRandom(new MersenneTwister());
                
                 CondorcetLearning lr = new CondorcetLearning(config.getInt("stratnum"));
                 
                 //Sets up the appropriate mutation rule.
                 MutateStrategy mr = new MutateStrategy(config.getDouble("mutationRate"), config.getInt("stratnum"));
                 if(config.contains("truerate")) {
                	 double fr;
                	 if(config.contains("falserate")) {
                		 fr = config.getDouble("falserate");
                	 }
                	 else {
                		 fr = config.getDouble("mutationRate");
                	 }
                		 
                	 double[][] mm = new double[config.getInt("stratnum")][config.getInt("stratnum")];
            	 	 mm[0][0] = (double) 1 - config.getDouble("truerate");
                	 for(int i=1;i<config.getInt("stratnum");i++) {
            	 		 mm[0][i] = (double) config.getDouble("truerate")/(config.getInt("stratnum")-1);
            	 	 }
                	 for(int i=1;i<config.getInt("stratnum");i++) {
                		 for(int j=0;j<config.getInt("stratnum");j++) {
                			 if(i == j) {
                				 mm[i][j] = (double) 1 - fr;
                			 }
                			 else {
                				 mm[i][j] = (double) fr/(config.getInt("stratnum") - 1);
                			 }
                		 }
                	 }
                 }
              
                 
                 /********************** Print out simulation parameters ***************/
                 System.out.println("Simple Condorcet Model parameter list:");
                 if(config.getBoolean("enumerate")) {
                	 System.out.println("   Enumeration");
                	 System.out.println("   Runs per job irrelevant");
                 }
                 else {
                	 System.out.println("   Simulation");
                	 System.out.println("   Runs per job: " + config.getLong("runs"));
                 }
                 System.out.println("   Job file: " + config.getString("jobFile"));
                 System.out.println("   Verbosity: " + config.getInt("verbose"));
                 System.out.println("   Max Generations: " + config.getInt("generations"));
                 System.out.println("   Strategies: " + config.getInt("stratnum") + " (0 is true)");
                 System.out.println("   Initial error rate: " + config.getDouble("epsilon"));
                 System.out.print("   Mutation rate: ");
                 if(config.contains("truerate")) {
                	 System.out.print(config.getDouble("truerate") + "(true -> non-true) " );
                	 if(config.contains("falserate")) {
                		 System.out.println(config.getDouble("falserate") + "(false -> other)");
                 	 }
                	 else {
                		 System.out.println(config.getDouble("mutationRate") + "(false -> other)");
                	 }
                		 
                 }
                 else {
                	 System.out.println(config.getDouble("mutationRate"));
                 }
                 
                 /******************* Simulation/Enumeration Section ********************/
                 
                 //Foreach job
                 while(todo.hasNext()) {
                	 
                	 Jobs job = (Jobs) todo.next();
                	 model.createPlayers(GraphIO.adjacencyMatrix(job.getMatrix()), mr, lr, new ConnectNothing());
                	 
                	 // If it's an enumeration
                	 
                	 if(config.contains("enumerate")) {
                		 int[] list = new int[model.getIndividuals().length];
                		 double probUnan = 0.0;
                		 double probMaj = 0.0;
                		 double probIndiv = 0.0;
                		 for(int i=0;i<list.length;i++) 
                			 list[i] = 0;
                		 while(list[0] != -1) {
                			 // Sets the initial strats
                			 for(int i=0;i<list.length;i++) {
                				 model.getIndividuals()[i].setStrat(list[i]);
                				 model.getIndividuals()[i].setLastStrat();
                			 }
                				
                			 
                			 int gens = model.runmodel(config.getDouble("mutationRate"),
                					 		config.contains("truerate"),
                					 		config.getInt("generations"),
                					 		config.getInt("verbose"));
                			 
                			 //caculates the probability of this state for record keeping purposes
                			 double probstate = 1.0;
                			 for(int i=0;i<model.getIndividuals().length;i++) {
                				 if(list[i] == 0) {
                					 probstate *= (double) 1 - config.getDouble("epsilon");
                				 }
                				 else {
                					 probstate *= (double) config.getDouble("epsilon")/(config.getInt("stratnum") - 1);
                				 }
                			 }

                			 probIndiv += (double) ((double) model.countTruth()/list.length) * probstate;
                			 
                			 switch(model.detectState()) {
            			 		case 2: probUnan += probstate;
            			 		case 1: probMaj += probstate;
                			 }
                			 
                			 if(config.getInt("verbose") == 1) {
                 				String output = new String(job.getJobID() + " " + job.getDescrip() + " " + probstate);
                 				for(int j=0;j<model.getIndividuals().length;j++) 
                 					output += " " + list[j];
                 				for(int j=0;j<model.getIndividuals().length;j++) 
                 					output += " " + model.getIndividuals()[j].getStrat();
                 				output += " " + model.detectState() + " " + model.countTruth() + " " + gens;
       
                 				System.out.println(output);
                 				
                 			}
                			 
                			 if(config.getInt("verbose") == 2) 
                				 System.out.println("** New Run **");
                			 
                			 list = model.incrementSequence(list, config.getInt("stratnum"));
                		 }
                	
                		 if(config.getInt("verbose") == 0) 
                			 System.out.println(job.getJobID() + " " + job.getDescrip() + " " + probUnan + " " + probMaj + " " + probIndiv);
                	 }
                	 
                	 // If it's a simulation 
                	 else {
                		long countUnan = 0L;
                		long countMaj = 0L;
                		long countIndiv = 0L;
                		for(long i=0;i<config.getLong("runs");i++) {
                			model.resetPlayers(config.getInt("stratnum"), config.getDouble("epsilon"));
                			
                			int[] firststate = new int[model.getIndividuals().length];
                			for(int j=0;j<model.getIndividuals().length;j++) 
                				firststate[j] = model.getIndividuals()[j].getStrat();
                			
                			int gens = model.runmodel(config.getDouble("mutationRate"),
        					 		config.contains("truerate"),
        					 		config.getInt("generations"),
        					 		config.getInt("verbose"));
                			
                			switch(config.getInt("verbose")) {
                			case 1:
                				String output = new String(job.getJobID() + " " + job.getDescrip());
                				for(int j=0;j<model.getIndividuals().length;j++) 
                					output += " " + firststate[j];
                				for(int j=0;j<model.getIndividuals().length;j++) 
                					output += " " + model.getIndividuals()[j].getStrat();
                				output += " " + model.detectState() + " " + gens;
                                      
                				System.out.println(output);
                				break;
                			case 2:
                				System.out.println("** New Run **");
                				break;
                			}
                			
                			countIndiv += model.countTruth();
                			
                			switch(model.detectState()) {
                				case 2: countUnan++;
                				case 1: countMaj++;
                			}
                		}
                		if(config.getInt("verbose") == 0) 
                			System.out.println(job.getJobID() + " " + 
                					job.getDescrip() + " " + 
                					((double) countUnan/config.getLong("runs")) + 
                					" " + ((double) countMaj/config.getLong("runs")) +
                					" " + ((double) countIndiv/(config.getLong("runs") * model.getIndividuals().length)));
                	 }
                	 
                	 
                 }
             }
             catch(ModelError e) {
                 System.out.println(e);
             }
             catch(FileNotFoundException e) {
            	 System.out.println(e);
             }
             catch(IOException e) {
                 System.out.println(e);
             }
             catch(JSAPException e) {
                 System.out.println(e);
             }
     }
}