package models;
import utilities.Game;
import utilities.Individual;
import utilities.MersenneTwister;
import utilities.ModelError;
import learningRules.BgLearning;
import learningRules.LearningRule;
import learningRules.SmoothBGLearning;
import learningRules.StratReinforce;
import mutationRules.MutateBeliefs;

import com.martiansoftware.jsap.*;

import connectionRules.ConnectNothing;
import connectionRules.ConnectionRule;

public class SingleBGM {
        
        Game[] games;
        
        /**
        * Sets an array of games, one for each state of the world.
        * @param g The array of Game objects
        */
        public synchronized void setGames(Game[] g) { games = g;}
        
        /**
        * Returns the games for the BgModel
        * @return the games
        */
        public Game[] getGames() { return games; }
        
          // The the individual
        Individual individual;
        /**
        * Sets the list of individuals for the model.
        * @param i An array of individuals
        */
        public synchronized void setIndividual(Individual i) { individual = i; }
        
        /**
        * Returns the set of individuals for the model.
        * @return the individuals
        */
        public Individual getIndividual() {return individual;}

        // The random number generator
        MersenneTwister random;
        /**
        * Sets the random number generator.
        * @param r the mersennetwister random number generator
        */
        public synchronized void setRandom(MersenneTwister r) { random =r; }
        
        /** 
        * Returns the random number generator for the model.
        * @return the mersenne twister random number generator
        */
        public MersenneTwister getRandom() { return random; }        
        
        /**
        * Runs the model
        *
        */
        public void run(JSAPResult config) throws ModelError{
                if(config.getInt("verbose")==0) {
                        System.out.println("WARNING: Don't use this verbosity level here.");
                }
                
                double avepayoff;
                long[] actioncount = new long[config.getInt("stratnum")];
                long[] statecount = new long[config.getInt("stratnum")];
                for(int i=0;i<config.getInt("stratnum");i++) 
                        statecount[i] = 0L;
                for(long runs=0L;runs<config.getLong("runs");runs++) {
                        int sow;
                        if(config.getBoolean("noBadState")) {
                                sow = random.nextInt(config.getInt("stratnum")-1) + 1;
                        }
                        else {
                                sow = random.nextInt(config.getInt("stratnum"));
                        }
                        statecount[sow]++;
                        individual.initStart(config.getInt("stratnum"));
                        individual.setGame(games[sow]);
                        
                        //set up statistics
                        avepayoff = 0.0;
                        for(int j=0;j<config.getInt("stratnum");j++) 
                                actioncount[j] = 0L;
                        
                        
                        int i;
                        for(i=0;i<config.getInt("generations");i++) {
                                if(config.getInt("verbose") == 2) {
                                        String beliefstring = "";
                                        if(config.getString("learningrule").equals("r")) {
                                                double total = 0.0;
                                                for(int j=0;j<config.getInt("stratnum");j++) 
                                                        total += individual.getWeights(j);
                                                for(int j=0;j<config.getInt("stratnum");j++) {
                                                        beliefstring += (double) individual.getWeights(j)/total;
                                                        beliefstring += " ";
                                                }
                                        }
                                        else {
                                                for(int j=0;j<config.getInt("stratnum");j++) {
                                                        beliefstring += individual.getBelief(j);
                                                        beliefstring += " ";
                                                }
                                        }
                                        System.out.print(runs + " " + sow + " " + i + " " + beliefstring);
                                        if(i==0) {
                                                System.out.println("- 0");
                                        }
                                        else {
                                                System.out.println(individual.getStrat() + " " + individual.getPayoff());
                                        }
                                                
                                }
                                
                                //Actually do the generations
                                individual.reset();
                                individual.playGame();
                                individual.update();
                                
                                // update statistics
                                
                                avepayoff += (double) individual.getPayoff()/config.getInt("generations");
                                actioncount[individual.getStrat()]++;
                        }
                        switch(config.getInt("verbose")) {
                                case 0: 
                                        if(runs%250==0) 
                                                System.out.print("*");
                                        break;
                                case 1: 
                                        String beliefstring = "";
                                        if(config.getString("learningrule").equals("re")) {
                                                double total = 0.0;
                                                for(int j=0;j<config.getInt("stratnum");j++) 
                                                        total += individual.getWeights()[j];
                                                for(int j=0;j<config.getInt("stratnum");j++) {
                                                        beliefstring += (double) individual.getWeights()[j]/total;
                                                        beliefstring += " ";
                                                }
                                        }
                                        else {
                                                for(int j=0;j<config.getInt("stratnum");j++) {
                                                        beliefstring += individual.getBelief(j);
                                                        beliefstring += " ";
                                                }
                                        }
                                        String actioncountstr = "";
                                        for(int j=0;j<config.getInt("stratnum");j++) { 
                                                actioncountstr += actioncount[j];
                                                actioncountstr += " ";
                                        }
                                        System.out.println(runs + " " + sow + " " 
                                                + avepayoff + " " + actioncountstr + beliefstring + i);
                                        break;
                                case 2:
                                        System.out.println("** New run **");
                                        break;
                        }               
                }
                if(config.getInt("verbose")==0) {
                        System.out.println("Since we're not doing convergence detection, you really can't learn anything at this verbosity level");
                }
        }
        
        public SingleBGM(JSAPResult config) throws ModelError {
                
                random = new MersenneTwister();
                
                //Set the zeroconstraint
                double zeroconstraint = 0.0;
                if(config.getBoolean("biasedPriors")) 
                        zeroconstraint = (double) (config.getInt("stratnum")-1)/config.getInt("stratnum");
                
                //Parameters of the model
                boolean stochastic = true;
                ConnectionRule c = new ConnectNothing();
                
                // The second argument gives the number of strategies, we will now generate a
                //   a game for that number of strategies.  If an agent assigns equal probability
                //   to all states, she is indifferent between the actions.
                
                int[][][] matrix = new int[config.getInt("stratnum")][config.getInt("stratnum")][];
                
                //Set up the bad payoff and good payoffs
                int badpayoff = config.getIntArray("payoffs")[0]-1;
                int goodpayoff = (config.getIntArray("payoffs")[0] * config.getInt("stratnum")) - (badpayoff*(config.getInt("stratnum")-1));
                int[] uninformmatrix = new int[config.getInt("stratnum")];
                int[] badmatrix = new int[config.getInt("stratnum")];
                int[] goodmatrix = new int[config.getInt("stratnum")];
                for(int i=0;i<config.getInt("stratnum");i++) {
                        uninformmatrix[i] = config.getIntArray("payoffs")[0];
                        badmatrix[i] = badpayoff;
                        goodmatrix[i] = goodpayoff;
                }
                
                //Set up the games
                games = new Game[config.getInt("stratnum")];
                for(int i=0;i<config.getInt("stratnum");i++) {
                        matrix[i][0] = uninformmatrix;
                        
                        for(int j=1;j<config.getInt("stratnum");j++) {
                                if(i==j) 
                                        matrix[i][j] = goodmatrix;
                                else
                                        matrix[i][j] = badmatrix;
                        }
                        games[i] = new Game(stochastic, random);
                        games[i].setMatrix(matrix[i], badmatrix[0] + goodmatrix[0]);
                        games[i].setStrategic(false);
                }
                
                // Set the learning rule
                LearningRule lr;
                if(config.getString("learningrule").equals("fp")) {
                        lr = new BgLearning(games);
                }
                else {
                        if(config.getString("learningrule").equals("sfp")) {
                                lr = new SmoothBGLearning(games, config.getDouble("gamma"));
                        }
                        else {
                                lr = new StratReinforce(0.01);
                        }
                }
                
                // Set up the individual
                individual = new Individual();
                Individual[] network = new Individual[1];
                network[0] = individual;
                individual.setNetwork(network);
                individual.setRandom(random);
                individual.setConnectionRule(c);
                individual.setLearningRule(lr);
                individual.setMutationRule(new MutateBeliefs(config.getDouble("mutationRate")));
                individual.setConstraint(zeroconstraint);
                individual.setMaxWeight(config.getInt("maxWeight"));
        }
}

