package utilities;
import java.util.*;
import connectionRules.ConnectionRule;
import learningRules.LearningRule;
import mutationRules.MutationRule;
import mutationRules.MutateBeliefs;

/**
* A class that implements individuals for agent based modeling.  This is compatible with any
* form of game and either reinforcement or bayesian learning.   Very little processing is
* done in this class.  Only some basic sanity checks for assignments.  In addition, I have
* removed functionality for keeping track of history of play beyond one round.  
* For learning rules which
* require longer memory of actions, one should implement another class.  (Keeping track of past
* play was very taxing on resources, and so was removed.)
* @author Kevin Zollman
*/

public class Individual {
        
        
        /**
        * An id number to identify individuals.  This number is not used for
        * anything internally, do with it what you like.
        */
        private int id;
        /**
        * Sets the id number.
        * @param i the id number
        */
        public synchronized void setId(int i) { id = i; }
        /**
        * Returns an the individuals id number
        * @return the id number
        */
        public int getId() { return id; }
        
/*------------------ Beliefs Section -------------------------------*/
        private double[][] beliefs;
        private double[] zeroconstraint;
        /**
        * A function to set the Bayesian beliefs for one domain.  This
        *   should be used for situations where the agent is only inferring about one
        *   state of the world.
        * @param b A 1-D double array containing beliefs.
        * @throws ModelError if there is something wrong with the beliefs
        */
        public synchronized void setBeliefs(double[] b) throws ModelError { 
                double check=0.0;
                for(int i=0;i<b.length;i++) 
                        check += b[i];
                if((check < 0.99999) || (check > 1.00001))
                        throw new ModelError("Belief "+check+" doesn't sum to 1.");                        
                beliefs = new double[1][];
                beliefs[0] = b; 
        }
        
        /**
        * A function to set the Bayesian beliefs for multi-domain situations.  This
        *   should be used for cases where the agent is trying to learn several 
        *   different things at once
        * @param b A 2-D double array containing beliefs.
        * @throws ModelError if there is something wrong with the beliefs
        */
        public synchronized void setBeliefs(double[][] b) throws ModelError { 
                for(int j=0;j<b.length;j++) {
                        double check=0.0;
                        for(int i=0;i<b[j].length;i++) 
                                check += b[j][i];
                        if((check < 0.99999) || (check > 1.00001))
                                throw new ModelError("Belief about " + j +": "+check+" doesn't sum to 1.");                        
                }
                beliefs = b; 
        }
        
        /**
        * A function to set the Bayesian beliefs for one domain in mutli-domain learning
        *   situations. This
        *   should be used for cases where the agent is trying to learn several 
        *   different things at once
        * @param b A 2-D double array containing beliefs.
        * @param d The domain to set
        * @throws ModelError if there is something wrong with the beliefs
        */
        public synchronized void setBeliefs(double[] b, int d) throws ModelError { 
                double check=0.0;
                for(int i=0;i<b.length;i++) 
                        check += b[i];
                if((check < 0.99999) || (check > 1.00001))
                        throw new ModelError("Belief about " + d +": "+check+" doesn't sum to 1.");                        
                int beliefSize = 1;
                if(d >= beliefs.length) 
                	beliefSize = d+1;
                else
                	beliefSize=beliefs.length;
                double[][] newbeliefs = new double[beliefSize][];
                for(int i=0;i<beliefSize;i++) {
                	if(i!=d) {
                		if(i>beliefs.length) {
                			newbeliefs[i] = new double[1];
                		}
                		else {
                			newbeliefs[i] = beliefs[i];
                		}	
                	}
                	else 
                		newbeliefs[i] = b;
                }
                beliefs = newbeliefs; 
        }
        
        /**
        * Returns an agent's beliefs.  Generally this shouldn't be used, prefer the specific
        *    ones.
        * @return The agent's beliefs
        */
        public double[][] getBeliefs() { return beliefs; }
        
        /**
        * Returns an agent's beliefs about a particular state in a model where the agent
        *    is only learning about one thing.
        * @param b the state we're interested in
        * @return the agent's belief in that state
        */
        public double getBelief(int b) { return beliefs[0][b]; }
        
        /**
        * Returns an agent's belief about a particular state in a model where there
        *   are many things about which the agent may be learning.
        * @param d the domain that the agent is learning about
        * @param b the thing about which where interested
        * @return the agent's belief about that state in that domain
        */
        public double getBelief(int d, int b) { return beliefs[d][b]; }
        
        /**
        * Sets a constraint that initializations can be constrained by.  The constraint 
        *   is a minimum for beliefs in state zero.  This should only be used for single
        *   domain learning.  Currently only zero state constraints are supported.
        * @param c A double representing the constraint
        * @throws ModelError if the constraint is outside of [0,1]
        */
        public synchronized void setConstraint(double c) throws ModelError {
                if((c>1.0)||(c<0.0)) 
                        throw new ModelError("Constraint outside of [0,1]");
                zeroconstraint = new double[1];
                zeroconstraint[0] = c;
        }
        
        /**
        * Sets a constraint that initializations can be constrained by.  The constraint 
        *   is a minimum for beliefs in state zero.  This function is for multiple 
        *   domain models. Currently only zero state constraints are supported.
        * @param c A double array representing the constraints
        * @throws ModelError if the constraint is outside of [0,1]
        */
        public synchronized void setConstraint(double[] c) throws ModelError {
                for(int i=0;i<c.length;i++) {
                        if((c[i]>1.0)||(c[i]<0.0)) 
                                throw new ModelError("Constraint outside of [0,1]");
                }
                zeroconstraint = c;
        }
        
        /**
        * Sets a constraint that initializations can be constrained by.  The constraint 
        *   is a minimum for beliefs in state zero.  This function is for multiple 
        *   domain models. Currently only zero state constraints are supported.  This
        *   function should only be used for reseting constraints once already set.
        * @param d the domain in which the constraint applies
        * @param c A double representing the constraint
        * @throws ModelError if the constraint is outside of [0,1]
        */
        public synchronized void setConstraint(int d, double c) throws ModelError {
                if((c>1.0)||(c<0.0)) 
                        throw new ModelError("Constraint outside of [0,1]");
                zeroconstraint[d] = c;
        }
        
        /**
        * Returns the constraints on generating new beliefs in a one domain model
        * @return the constraints
        */
        public double getConstraint() { return zeroconstraint[0]; }
        
        
        /**
        * Returns the constraints on generating new beliefs for a particular domain.  
        * @param d the domain for the constraint
        * @return the constraints
        */
        public double getConstraint(int d) { return zeroconstraint[d]; }
        
        /** 
        * Initializes the beliefs for a multi-domain model.  This muse be called
        *   prior to accessing beliefs in a particular domain.
        * @param d the number of domains in the model
        */
        public void initDomains(int d) { beliefs = new double[d][]; }
        
        /** 
        * Randomizes an agents beliefs, this requires that the random be set.  The randomized
        *     beliefs will be on the interior of [0,1].  It is generated by the 
        *     logrithm method.  This is for one domain models.
        * @param states the number of states over which to generate beliefs
        * @param constrained a boolean to decide if we'll pay attention to the constraint
        */
        public synchronized void randBeliefs(int states, boolean constrained) {
                double number[] = new double[states];
                beliefs = new double[1][];
                
                boolean interior;
                do {
                interior = true;

                double total = 0.0;
                for(int i=0;i<states;i++) {
                        number[i] = (double) -1*Math.log(random.nextDouble());
                        total += number[i];
                }
                
                beliefs[0] = new double[states];
                for(int i=0;i<states;i++)  {
                        beliefs[0][i] = (double) number[i]/total;
                        if(beliefs[0][i] == 1 || beliefs[0][i] == 0 
                                || Double.isNaN(beliefs[0][i]) 
                                || beliefs[0][i] == Double.POSITIVE_INFINITY ||
                                beliefs[0][i] == Double.NEGATIVE_INFINITY) {
                                interior=false;
                                break;
                        }
                }
                } while(!interior ||  (constrained && (beliefs[0][0] < zeroconstraint[0]) ));
        }
        
        /** 
        * Randomizes an agents beliefs, this requires that the random be set.  The randomized
        *     beliefs will be on the interior of [0,1].  It is generated by the 
        *     logrithm method.  This function only initializes for a single domain, and 
        *     it requires that initDomain() be called first.
        * @param states the number of states over which to generate beliefs
        * @param d which domain should be randomly assigned
        * @param constrained a boolean to decide if we'll pay attention to the constraint
        */
        public synchronized void randBeliefs(int states, int d, boolean constrained) {
                double number[] = new double[states];
                
                boolean interior;
                do {
                interior = true;
                double total = 0.0;
                for(int i=0;i<states;i++) {
                        number[i] = (double) -1*Math.log(random.nextDouble());
                        total += number[i];
                }
                
                beliefs[d] = new double[states];
                for(int i=0;i<states;i++)  {
                        beliefs[d][i] = (double) number[i]/total;
                        if(beliefs[d][i] == 1 || beliefs[d][i] == 0 
                                || Double.isNaN(beliefs[d][i]) 
                                || beliefs[d][i] == Double.POSITIVE_INFINITY ||
                                beliefs[d][i] == Double.NEGATIVE_INFINITY) {
                                interior=false;
                                break;
                        }
                }
                } while(!interior ||  (constrained && (beliefs[d][0] < zeroconstraint[d]) ));
        }
        
        /**
        * Randomizes an agents beliefs, this requires that the random be set.  The randomized
        * beliefs will be on the interior of [0,1].  It is generated by the logrithm method. 
        * This function pays attention to the constraint by default.
        * @param states the number of states over which to generate beliefs
        * 
        */
        public synchronized void randBeliefs(int states) { randBeliefs(states, true); }
        
        /**
        * Randomizes an agents beliefs, this requires that the random be set.  The randomized
        * beliefs will be on the interior of [0,1].  It is generated by the logrithm method. 
        * This function pays attention to the constraint by default.
        * @param states the number of states over which to generate beliefs
        * @param d the domain in which the beliefs should be generated
        */
        public synchronized void randBeliefs(int states, int d) { randBeliefs(states,d,true); }
        
        /**
        * Generates uniform beliefs over a number of states.  Used for one domain models
        *    this function ignores the zero constraint, of course
        * @param states the number of states
        */
        public synchronized void uniformBeliefs(int states) {
                beliefs = new double[1][states];
                for(int i=0;i<states;i++) 
                        beliefs[0][i] = (double) 1/states;
        }
        
        /**
        * Generates uniform beliefs over a number of states.  Used for multi domain models
        *    Init domains must be called first.  This function ignores the zero constraint,
        *    of course
        * @param states the number of states
        * @param d the domain in which to create uniform beliefs
        */
        public synchronized void uniformBeliefs(int states, int d) {
                beliefs[d] = new double[states];
                for(int i=0;i<states;i++) 
                        beliefs[d][i] = (double) 1/states;
        }
        
        
/*------------------ Weights Section -------------------------------*/
        
        private double[][] stratWeights;
        private int maxWeight;
        
        /**
        * Sets the maxWeight
        * @param m The max weight
        */
        public synchronized void setMaxWeight(int m) { maxWeight = m; }
        
        /**
        * Gets the maxWeight
        * @return The max weight
        */
        public int getMaxWeight() { return maxWeight; }
        
        /**
        * A function to reset all strategy weights to 1.0.  This function presumes that strat
        * wieghts have already once been set, otherwise an error is thrown.
        */
        public synchronized void clearStratWeights() {
                for(int i=0; i<stratWeights.length; i++) {
                        for(int j=0; j<stratWeights[i].length; j++) 
                                stratWeights[i][i] = 1.0;
                }
        }
        
        /**
        * A function that will clear the weights in only one domain, but leave the others 
        * intact
        * @param d The domain to clear
        */
        public synchronized void clearDomainWeights(int d) {
                for(int i=0; i<stratWeights.length; i++)
                        stratWeights[d][i] = 1.0;
        }
        
        /**
        * A function to reset all strategy weights to 1.0.  This function does not presume that
        * strategy weights are already set.  This is for only a single domain.
        * @param n the number of states which require weights
        */
        public synchronized void clearStratWeights(int n) {
                stratWeights = new double[1][];
                stratWeights[0] = new double[n];
                for(int i=0; i<n; i++)
                        stratWeights[0][i] = 1.0;
        }
        
        /**
        * A function to reset all stategy weights to 1.0.  This function does not presume that
        * strategy weights have already been set.  This for multi-domain weights
        * @param n the number of strategies that require weights
        * @param d the number of domains
        */
        public synchronized void clearStratWeights(int n, int d) {
                stratWeights = new double[d][];
                for(int i=0; i<d; i++) {
                        stratWeights[i] = new double[n];
                        for (int j=0;j<n;j++)                        
                                stratWeights[i][j] = 1.0;
                }
        }
        
        /**
        * Randomizes the weights using the set maximum.  This function presumes that
        * random has been set.  This is for single domain models.
        * @param n The number of strategies
        */
        public synchronized void randStratWeights(int n) {
                stratWeights = new double[1][];
                stratWeights[0] = new double[n];
                for(int i=0;i<n;i++)
                        stratWeights[0][i] = random.nextDouble() + random.nextInt(maxWeight);
        }
        
        /**
        * Randomizes the weights using the set maximum.  This function presumes that
        * random has been set.  This is for single domain models.
        * @param n The number of strategies
        */
        public synchronized void randStratWeightsD(int n, int d) {
                stratWeights = new double[d][];
                for(int i=0;i<d;i++) {
                        stratWeights[i] = new double[n];
                        for(int j=0;j<n;j++) 
                                stratWeights[i][j] = random.nextDouble() + random.nextInt(maxWeight);
                }
        }
        
        /**
        * Sets the strategy weights to a specific value.  This function will return an error if
        * any of the weights are negative.  0 weights are allowed, although keep in mind this
        * means that that strategy will never be selected.
        * @param s the weights
        * @throws ModelError if any weight is negative
        */
        public synchronized void setStratWeights(double s[]) throws ModelError { 
                for(int i=0;i<s.length;i++) {
                        if(s[i] < 0)
                                throw new ModelError("StratWeight negative");
                }
                stratWeights[0] = s; 
        }
        
        /**
        * Sets the strategy weights to a specific value.  This function will return an error if
        * any of the weights are negative.  0 weights are allowed, although keep in mind this
        * means that that strategy will never be selected.
        * @param s the weights
        * @throws ModelError if any weight is negative
        */
        public synchronized void setStratWeights(double s[][]) throws ModelError {
                for(int i=0;i<s.length;i++) {
                        for(int j=0;j<s[i].length;j++) {
                                if(s[i][j] < 0)
                                        throw new ModelError("StratWeight negative");
                        }
                }
                stratWeights = s; 
        }
        
        /**
        * Sets the strategy weight for a particular strategy to a specific value. This function will return an error if
        * any of the weights are negative.  0 weights are allowed, although keep in mind this
        * means that that strategy will never be selected.
        * @param p the strategy whose weight we will change
        * @param s the new weight
        * @throws ModelError if the weight is less than 0
        */
        public synchronized void setStratWeights(int p, double s) throws ModelError {
                if(s < 0)
                        throw new ModelError("StratWeight negative");
                stratWeights[0][p] = s;
        }
        
        /**
        * Sets the strategy weight for a particular strategy to a specific value. This function will return an error if
        * any of the weights are negative.  0 weights are allowed, although keep in mind this
        * means that that strategy will never be selected.
        * @param d the domain
        * @param p the strategy whose weight we will change
        * @param s the new weight
        * @throws ModelError if the weight is less than 0
        */
        public synchronized void setStratWeights(int d, int p, double s) throws ModelError {
                if(s < 0)
                        throw new ModelError("StratWeight negative");
                stratWeights[d][p] = s;
        }
        
        /**
        * Adds a specified weight to a particular strategy.  This function will
        * return an error if the weight is negative.
        * @param p the strategy to update
        * @param w the amount to add
        * @throws ModelError if the number to add is negative
        */
        public synchronized void addStratWeight(int p, double w) throws ModelError { 
                if(w < 0)
                        throw new ModelError("Added weight negative");
                stratWeights[0][p] += w; 
        }
        
        /**
        * Adds a specified weight to a particular strategy.  This function will
        * return an error if the weight is negative.
        * @param d the domain
        * @param p the strategy to update
        * @param w the amount to add
        * @throws ModelError if the number to add is negative
        */
        public synchronized void addStratWeight(int d, int p, double w) throws ModelError { 
                if(w < 0)
                        throw new ModelError("Added weight negative");
                stratWeights[d][p] += w; 
        }
        
        /**
        * Returns the current weights. 
        * @return a double array which are the agents weights
        */
        public double[] getWeights() { return stratWeights[0]; }
        
        /**
        * Returns the weights for a current strategy
        * @param p the strategy for which we want weights
        * @return the agent's weight for that strategy
        */
        public double getWeights(int p) {return stratWeights[0][p]; }
        
        /**
        * Returns the weights for a current strategy
        * @param d the domain weights.
        * @param p the strategy for which we want weights
        * @return the agent's weight for that strategy
        */
        public double getWeights(int d, int p) {return stratWeights[d][p]; }
        
        /** 
        * Returns the full set of weights for a particular domain
        * @param d the domain
        */
        public double[] getDomainWeights(int d) {return stratWeights[d]; }
        
        
        // Discount over time
        private double discount;
        
        /**
        * Sets a discount value
        * @param d the discount factor
        * @throws ModelError if the discount factor is outside [0,1]
        */
        public synchronized void setDiscount(double d) throws ModelError { 
                if((d<0) || (d>1)) 
                        throw new ModelError("Discount not [0,1]");
                discount = d; 
        }
        
        /**
        * Returns the discount factor
        * @return the discount factor
        */
        public double getDiscount() { return discount; }
        
        
/*------------------ Alpha/Beta Beliefs Section -------------------------------*/
        double[] alpha;
        double[] beta;
        
        int maxalpha;	
        int maxbeta;
        /**
	 * Returns the value of maxalpha used to contrain the reseting of alpha.
	 */
	public int getMaxalpha() { return maxalpha;	}

	/**
	 * Sets the value of maxalpha.
	 * @param maxalpha The value to assign maxalpha.
	 */
	public void setMaxalpha(int maxalpha) { this.maxalpha = maxalpha; }

        /**
	 * Returns the value of maxbeta.
	 */
	public int getMaxbeta() { return maxbeta; }

	/**
	 * Sets the value of maxbeta.
	 * @param maxbeta The value to assign maxbeta.
	 */
	public void setMaxbeta(int maxbeta) { this.maxbeta = maxbeta; }
        
        /**
        * Gets the alpha value
        * @return alpha
        */
        public double[] getAlpha() { return alpha; }
        
        /**
        * Gets the beta value
        * @return beta
        */
        public double[] getBeta() { return beta; }
        
        /**
        * Gets the alpha value for a particular domain
        * @param d the domain
        * @return alpha
        */
        public double getAlpha(int d) { return alpha[d]; }
        
        /**
        * Gets the beta value for a particular domain
        * @param d the domain
        * @return beta
        */
        public double getBeta(int d) { return beta[d]; }
        
        /**
        * Adds a value to the alpha function, used in updating
        * @param a amount to add to alpha
        * @param d the domain to which to add
        */
        public void addAlpha(double a, int d) { alpha[d] += a; } 
        
        /**
        * Adds a value to the alpha function, used in updating
        * @param a amount to add to alpha
        * @param d the domain to which to add
        */
        public void addBeta(double a, int d) { beta[d] += a; } 
        
        /**
        * Sets alpha at a specific value
        * @param a the value to set alpha
        * @param d the domain
        */ 
        public void setAlpha(double a, int d) { alpha[d] = a; }
        
        /**
        * Sets beta at a specific value
        * @param b the value to set beta
        * @param d the domain
        */ 
        public void setBeta(double b, int d) { beta[d] = b; }
        
        /**
        * Sets alpha at a specific value
        * @param a the value to set alpha
        */ 
        public void setAlpha(double a[]) { alpha = a; }
        
        /**
        * Sets beta at a specific value
        * @param b the value to set beta
        */ 
        public void setBeta(double b[]) { beta = b; }
        
        /**
        * Randomizes alpha and beta.  Numbers are chosen between 0 and the specified
        *   upper bound.  Sets maxalpha and maxbeta to a and b.  Requires random to be set.
        * @param a Upper bound for alpha
        * @param b Upper bound for beta
        * @param d The number of domains
        */
        public void randomAB(int a, int b, int d) {
                maxalpha = a; 
                maxbeta = b;
                alpha = new double[d];
                beta = new double[d];
                for(int i=0;i<d;i++) {
                        alpha[i] = (double) random.nextDouble() + random.nextInt(a);
                        beta[i] = (double) random.nextDouble() + random.nextInt(b);
                }
        }
        
        /**
        * Randomizes alpha and beta.  Numbers are chosen between 0 and the specified
        *   upper bound.  Uses maxalpha and maxbeta which must already be set
        *   Requires random to be set.
        * @param d The number of domains
        */
        public void randomAB(int d) {
                alpha = new double[d];
                beta = new double[d];
                for(int i=0;i<d;i++) {
                        alpha[i] = (double) random.nextDouble() + random.nextInt(maxalpha);
                        beta[i] = (double) random.nextDouble() + random.nextInt(maxbeta);
                }
        }
        
        /**
        * Randomizes alpha and beta.  Numbers are chosen between 0 and the specified
        *   upper bound.  Uses maxalpha and maxbeta which must already be set, and resets
        *   for the same size domain.  Requires random to be set.
        * @param d The number of domains
        */
        public void randomAB() {
                for(int i=0;i<alpha.length;i++) {
                        alpha[i] = (double) random.nextDouble() + random.nextInt(maxalpha);
                        beta[i] = (double) random.nextDouble() + random.nextInt(maxbeta);                }
        }
        
        /**
        * Initializes the alpha's without setting them.
        * @param d the number of domains
        */
        public void initAB(int d) { alpha = new double[d]; beta = new double[d]; }
        
/*------------------ Mutation Section -------------------------------*/
        private MutationRule mutationRule;   
               
        public void setMutationRule(MutationRule m) {
        	mutationRule = m;
        }
        
        public MutationRule getMutationRule() { return mutationRule; }
        
        /**
        * Calls the mutationRule set function.  
        * @param m the mutation rate
        * @throws ModelError if the mutation rate is outside of [0,1]
        */
        public synchronized void setMutation(double m) throws ModelError { 
                mutationRule.setMutationRate(m, game.getStrats());
        }
        
        /**
        * Calls the mutationRules get function.
        * @return the mutation rate
        */
        public double getMutation() { return mutationRule.getMutationRate(); }
        
        /**
        * Calls the mutationRules mutate function.
        */
        public void mutate() throws ModelError { mutationRule.mutate(this); }
        
/*------------------ Networks Section -------------------------------*/
        
        // A reference to the entire network of individuals (necessary for dynamic nets)
        private Individual[] network;
        
        /**
        * Sets the network (all other individuals in the model).
        * @param nw an array of Individuals that constitutes the network
        */
        public synchronized void setNetwork(Individual[] nw) { network = nw; }
        
        /**
        * Returns the network (all other individuals in the model).
        * @return the network
        */
        public Individual[] getNetwork() { return network; }

        // An arraylist of individuals to whom our player is connected.
        private HashSet connected;
        /**
        * Clears the list of neighbors.
        */
        public synchronized void clearConnected() { connected = new HashSet(); }
        
        /**
        * Sets the neighbors to a particular HashSet of individuals
        * @param c The new neighbors
        */
        public synchronized void setConnected(HashSet c) { connected = c; }        
        
        /**
        * Adds a neighbor to an individuals current neighbors.
        * @param i The individual to add to the list of neighbors
        */
        public synchronized void addConnected(Individual i) { connected.add(i); }
        
        /**
        * Returns the individuals neighbors
        * @return A HashSet of Individuals
        */
        public HashSet getConnected() { return connected; }

        
/*------------------ Random Number Section -------------------------------*/
        // Ye ole random number generator
        private MersenneTwister random;
        
        /**
        * Sets the random number generator
        * @param r The MersenneTwister random number generator
        */
        public synchronized void setRandom(MersenneTwister r) { random = r; }
        
        /**
        * Returns the random number generator
        * @return The MersenneTwister random number generator
        */
        public MersenneTwister getRandom() { return random; }
        
/*------------------ Learning Rules Section -------------------------------*/
        
        // The learning rule for forming new connections
        private ConnectionRule connectionRule;
        
        /**
        * Sets the connection rule, which determines how the individual updates its neighborhood
        * @param c The connection rule
        */
        public synchronized void setConnectionRule(ConnectionRule c) { connectionRule = c; }
        
        /**
        * Uses the connection rule to update the neighborhood.
        */
        public synchronized void formConnections() { connected = connectionRule.formConnection(this); }
        
        /**
        * Returns the connection rule
        * @return The connection rule
        */
        public ConnectionRule getConnectionRule() { return connectionRule; }
        
        // The learning rule for strategies
        private LearningRule learningRule;
        
        /**
        * Sets the learning rule for updating strategies
        * @param l The learning rule to set
        */
        public synchronized void setLearningRule(LearningRule l) { learningRule = l; }
        
        /**
        * Returns the learning rule
        * @return The learning rule
        */
        public LearningRule getLearningRule() {return learningRule;}
        
        
/*------------------ Strategies Section -------------------------------*/
        // The strategy to be used for this generation
        private int strat;
        
        /**
        * Sets the current strategy of a user
        * @param s The strategy
        * @throws ModelError if the strategy is negative or bigger than the possible number of 
        *                    strats
        */
        public synchronized void setStrat(int s) throws ModelError {
                if(s<0) 
                        throw new ModelError("Strategy less than zero");
                strat = s; 
        }
        
        /**
        * Uses the learning rule to choose a strategy
        * @return the strategy just selected
        * @throws ModelError if there is a problem selecting the strategy
        */
        public synchronized int chooseStrat() throws ModelError { 
                setStrat(learningRule.pickStrat(this)); 
                return strat; 
        }
        
        /**
        * Returns the current strategy, if its not set chooses one.
        * @return the strategy
        * @throws ModelError if there is a problem choosing the strategy.
        */
        public int getStrat() throws ModelError { if(strat == -1) setStrat(learningRule.pickStrat(this)); return strat; }
        
        // This is the strategy used on the last generation
        private int lastroundstrat;
        
        /**
        * This returns the strategy used on the last round
        * @return The strategy
        */
        public int getLastStrat() { return lastroundstrat; }
        
        /**
        * This updates the strategy for the next round by setting the last round
        * strat equal to the current strat.
        */
        
        public void setLastStrat() { lastroundstrat = strat; }
        
        // Payoff for generation
        private int payoff;
        
        /**
        * Returns the payoff from the last round
        * @return the payoff
        */
        public int getPayoff() { return payoff; }
        
        /**
        * Sets the payoff from the last generation
        * @param p the payoff
        */
        public synchronized void setPayoff(int p) { payoff = p; }
        
        /** 
        * Adds a value to the current payoff
        * @param p The amount to add
        */
        public synchronized void addPayoff(int p) { payoff += p; }
        
        private int lastroundpayoff;
        
        /** 
        * Returns the payoff from the last round
        */
        public int getLastPayoff() { return lastroundpayoff; }
        
        /**
        * Sets the last round payoff equal to the current payoff
        */
        public void setLastPayoff() { lastroundpayoff = payoff; }
        
        // The game being played
        private Game game;
        
        /**
        * Sets the game
        * @param g The game
        */
        public synchronized void setGame(Game g) { game = g; }
        
        /**
        * Returns the game
        * @return The game
        */
        public Game getGame() { return game; }
        
        /**
        * Resets for a new generation.  This runs the new generation function of the connection
        * and learning rules, chooses a new strategy, forms new connections, trys to mutate, and
        * sets the payoff to zero.
        * @throws ModelError If there is some problem
        */
        public synchronized void reset() throws ModelError {
                this.setLastStrat();
                this.setLastPayoff();
                connectionRule.newGeneration();
                learningRule.newGeneration(this);
                this.chooseStrat();
                this.formConnections();
                this.mutate();
                payoff = 0;
        }
        
        /**
        * Plays the game with each neighbor.  If the game is marked as strategic, this goes
        * through each neighbor and palys the game adding the payoff to the total payoff.
        * Otherwise, plays the game once and adds this to the payoff.
        * @throws ModelError If there are any problems
        */
        public synchronized void playGame() throws ModelError {
                if(game.isStrategic() == true) {
                        Iterator i = connected.iterator();
                        while(i.hasNext()) {
                                Individual n = (Individual) i.next();
                                payoff += game.getPayoff(this.getStrat(), n.getStrat());
                        }
                }
                else {
                        payoff += game.getPayoff(this.getStrat());
                }
        }
        
        // Update after play
        /**
        * Updates after play is complete.  This only calls the learningrule process payoff 
        * function
        * @throws ModelError From processpayoff
        */
        public synchronized void update() throws ModelError {
                learningRule.processPayoff(this);
        }
        
        
        /**
        * The initializes the individual for single domain models by giving them random 
        *  beliefs and randomizing the weights
        * The random beliefs are constrained by the zeroconstraint. 
        * @param s The number of strategies
        * @throws ModelError If there are any problems
        */
        public synchronized void initStart(int s) throws ModelError {
                randBeliefs(s);
                randStratWeights(s);
                randomAB(s);
        }
        
        
        /**
        * The initializes the individual for multi domain models by giving them random beliefs 
        *  and randomizes the the wieghts
        * The random beliefs are constrained by the zeroconstraint. BTW, this is not configured
        * for multidomain weight functions, only beleifs.
        * @param s The number of strategies
        * @throws ModelError If there are any problems
        */
        public synchronized void initStart(int d, int s) throws ModelError {
                beliefs = new double[d][];
                for(int i=0;i<d;i++) 
                        randBeliefs(s, i);
                randStratWeights(s);
                randomAB(s);
        }
       
        
/*------------------ Constructors -------------------------------*/
        /**
        * Returns and individual object.  Sets up an empty list for neighbors, the strategy to -1,
        * the payoff to zero, the discount factor to 1.0 (no discounting), the zero constraint
        * to 0.0, maxalpha and maxbeta to 4, and maxWeights to 1.
        */
        public Individual() { connected = new HashSet(); strat = -1; 
                payoff = 0; discount = 1.0; zeroconstraint=new double[1];
                zeroconstraint[0] = 0.0; maxalpha = 4; maxbeta = 4; maxWeight = 1;
                lastroundstrat = -1; 
                try { mutationRule = new MutateBeliefs(0.0);}
                catch(ModelError e) { System.out.println(e); }
        }
        
        /**
        * Returns and individual object like the simple constructor, but also sets an id.
        * @param i The id
        */
        public Individual(int i) { this(); id=i; }
        
        /**
        * Returns and individual object, but with more initialized.
        * @param i The id
        * @param s The number of strategies
        * @param r The random number generator
        * @param nw The network
        * @param c The connection rule
        * @param l The learning rule
        */
        public Individual(int i, int s, MersenneTwister r,  Individual[] nw, ConnectionRule c, LearningRule l){
                this();
                id = i; stratWeights = new double[1][]; stratWeights[0] = new double[s];
                random = r; network = nw; connectionRule = c;
                learningRule =l;
                for(int j=0;j<s;j++) 
                        stratWeights[0][j] = 1.0;
        }
}
